# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import numpy as np
import sys

import time
import traceback
from sklearn.model_selection import train_test_split
from sklearn.model_selection._split import check_cv
from sklearn.metrics import log_loss
from sklearn.metrics.scorer import check_scoring

# Import ConfigSpace and different types of parameters
from smac.configspace import ConfigurationSpace

# Import SMAC-utilities
from smac.facade.smac_facade import SMAC as orig_SMAC
from smac.scenario.scenario import Scenario
from smac.tae.execute_ta_run import BudgetExhaustedException
from lale.helpers import cross_val_score_track_trials
from lale.lib.sklearn import LogisticRegression
import lale.operators
from lale.search.lale_smac import lale_op_smac_tae, get_smac_space, lale_trainable_op_from_config
import lale.sklearn_compat
import lale.docstrings

logger = logging.getLogger(__name__)

class SMACImpl:

    def __init__(self, estimator=None, max_evals=50, cv=5, handle_cv_failure=False, scoring='accuracy', best_score=0.0, max_opt_time=None, lale_num_grids=None):
        """ Instantiate the SMAC that will use the given estimator and other parameters to select the 
        best performing trainable instantiation of the estimator. 

        Parameters
        ----------
        estimator : lale.operators.IndividualOp or lale.operators.Pipeline, optional
            A valid Lale individual operator or pipeline, by default LogisticRegression
        max_evals : int, optional
            Number of trials of SMAC search i.e. runcount_limit of SMAC, by default 50
        cv : an integer or an object that has a split function as a generator yielding (train, test) splits as arrays of indices.
            Integer value is used as number of folds in sklearn.model_selection.StratifiedKFold, default is 5.
            Note that any of the iterators from https://scikit-learn.org/stable/modules/cross_validation.html#cross-validation-iterators can be used here.
            The fit method performs cross validation on the input dataset for per trial, 
            and uses the mean cross validation performance for optimization. This behavior is also impacted by handle_cv_failure flag, 
            by default 5
        handle_cv_failure : bool, optional
            A boolean flag to indicating how to deal with cross validation failure for a trial.
            If True, the trial is continued by doing a 80-20 percent train-validation split of the dataset input to fit
            and reporting the score on the validation part.
            If False, the trial is terminated by assigning status to FAIL.
            , by default False
        scoring: string or a scorer object created using 
            https://scikit-learn.org/stable/modules/generated/sklearn.metrics.make_scorer.html#sklearn.metrics.make_scorer.
            A string from sklearn.metrics.SCORERS.keys() can be used or a scorer created from one of 
            sklearn.metrics (https://scikit-learn.org/stable/modules/classes.html#module-sklearn.metrics).
            A completely custom scorer object can be created from a python function following the example at 
            https://scikit-learn.org/stable/modules/model_evaluation.html
            The metric has to return a scalar value, and note that scikit-learns's scorer object always returns values such that
            higher score is better. Since Hyperopt solves a minimization problem, we pass (best_score - score) to Hyperopt.
            by default 'accuracy'.
        best_score : float, optional
            The best score for the specified scorer. This allows us to return a loss to hyperopt that is
            greater than equal to zero, where zero is the best loss. By default, zero.
        max_opt_time : float, optional
            Maximum amount of wall clock time in seconds for the optimization. By default, None, implying no runtime
            bound.

        Examples
        --------
        >>> from sklearn.metrics import make_scorer, f1_score, accuracy_score
        >>> lr = LogisticRegression()
        >>> clf = SMAC(estimator=lr, scoring='accuracy', cv=5)
        >>> from sklearn import datasets
        >>> diabetes = datasets.load_diabetes()
        >>> X = diabetes.data[:150]
        >>> y = diabetes.target[:150]
        >>> trained = clf.fit(X, y)
        >>> predictions = trained.predict(X)

        Other scoring metrics:

        >>> clf = SMAC(estimator=lr, scoring=make_scorer(f1_score, average='macro'), cv=3, max_evals=2)

        """

        self.max_evals = max_evals
        if estimator is None:
            self.estimator = LogisticRegression()
        else:
            self.estimator = estimator

        self.search_space:ConfigurationSpace = get_smac_space(self.estimator, lale_num_grids=lale_num_grids)
        self.scoring = scoring
        self.best_score = best_score
        self.handle_cv_failure = handle_cv_failure
        self.cv = cv
        self.max_opt_time = max_opt_time
        # Scenario object
        scenario_options = {"run_obj": "quality",   # we optimize quality (alternatively runtime)
                            "runcount-limit": self.max_evals,  # maximum function evaluations
                            "cs": self.search_space,               # configuration space
                            "deterministic": "true",
                            "abort_on_first_run_crash": False,
                            }
        if max_opt_time is not None:
            scenario_options["wallclock_limit"]= max_opt_time
        self.scenario = Scenario(scenario_options)
        self.trials = None

    def fit(self, X_train, y_train):
        self.cv = check_cv(self.cv, y = y_train, classifier=True) #TODO: Replace the classifier flag value by using tags?

        def smac_train_test(trainable, X_train, y_train):
            try:
                cv_score, logloss, execution_time = cross_val_score_track_trials(trainable, X_train, y_train, cv=self.cv, scoring=self.scoring)
                logger.debug("Successful trial of SMAC")
            except BaseException as e:
                #If there is any error in cross validation, use the score based on a random train-test split as the evaluation criterion
                if self.handle_cv_failure:
                    X_train_part, X_validation, y_train_part, y_validation = train_test_split(X_train, y_train, test_size=0.20)
                    start = time.time()
                    trained = trainable.fit(X_train_part, y_train_part)
                    scorer = check_scoring(trainable, scoring=self.scoring)
                    cv_score  = scorer(trained, X_validation, y_validation)
                    execution_time = time.time() - start
                    y_pred_proba = trained.predict_proba(X_validation)
                    try:
                        logloss = log_loss(y_true=y_validation, y_pred=y_pred_proba)
                    except BaseException:
                        logloss = 0
                        logger.debug("Warning, log loss cannot be computed")
                else:
                    logger.debug("Error {} with pipeline:{}".format(e, trainable.to_json()))
                    raise e
            return cv_score, logloss, execution_time

        def f(trainable):
            return_dict = {}
            try:
                score, logloss, execution_time = smac_train_test(trainable, X_train=X_train, y_train=y_train)
                return_dict = {
                    'loss': self.best_score - score,
                    'time': execution_time,
                    'log_loss': logloss
                }
            except BaseException as e:
                logger.warning(f"Exception caught in SMACCV:{type(e)}, {traceback.format_exc()}, SMAC will set a cost_for_crash to MAXINT.")
                raise e
            return return_dict['loss']

        try :
            smac = orig_SMAC(scenario=self.scenario, rng=np.random.RandomState(42),
                    tae_runner=lale_op_smac_tae(self.estimator, f))
            incumbent = smac.optimize()
            self.trials = smac.get_runhistory()
            trainable = lale_trainable_op_from_config(self.estimator, incumbent)
            #get the trainable corresponding to the best params and train it on the entire training dataset.
            trained = trainable.fit(X_train, y_train)
            self._best_estimator = trained
        except BudgetExhaustedException:
            logger.warning('Maximum alloted optimization time exceeded. Optimization exited prematurely')
        except BaseException as e:
            logger.warning('Error during optimization: {}'.format(e))
            self._best_estimator = None

        return self

    def predict(self, X_eval):
        import warnings
        warnings.filterwarnings("ignore")
        trained = self._best_estimator
        try:
            predictions = trained.predict(X_eval)
        except ValueError as e:
            logger.warning("ValueError in predicting using SMACCV:{}, the error is:{}".format(trained, e))
            predictions = None

        return predictions

    def get_trials(self):
        """Returns the trials i.e. RunHistory object.
        
        Returns
        -------
        smac.runhistory.runhistory.RunHistory
            RunHistory of all the trials executed during the optimization i.e. fit method of SMACCV.
        """
        return self.trials

    def get_pipeline(self, pipeline_name=None, astype='lale'):
        if pipeline_name is not None:
            raise NotImplementedError('Cannot get pipeline by name yet.')
        result = getattr(self, '_best_estimator', None)
        if result is None or astype == 'lale':
            return result
        assert astype == 'sklearn', astype
        return lale.sklearn_compat.make_sklearn_compat(result)

_hyperparams_schema = {
    'allOf': [
    {   'type': 'object',
        'required': [
            'estimator', 'max_evals', 'cv', 'handle_cv_failure',
            'max_opt_time', 'lale_num_grids'],
        'relevantToOptimizer': ['estimator'],
        'additionalProperties': False,
        'properties': {
            'estimator': {
                'anyOf': [
                {   'laleType': 'operator',
                    'not': {'enum': [None]}},
                {   'enum': [None]}],
                'default': None},
            'max_evals': {
                'type': 'integer',
                'minimum': 1,
                'default': 50},
            'cv': {
                'type': 'integer',
                'minimum': 1,
                'default': 5},
            'handle_cv_failure': {
                'type': 'boolean',
                'default': False},
            'scoring': {
                'anyOf': [
                {    'description': 'Custom scorer object, see https://scikit-learn.org/stable/modules/model_evaluation.html',
                     'not': {'type': 'string'}},
                {    'enum': [
                        'accuracy', 'explained_variance', 'max_error',
                        'roc_auc', 'roc_auc_ovr', 'roc_auc_ovo',
                        'roc_auc_ovr_weighted', 'roc_auc_ovo_weighted',
                        'balanced_accuracy', 'average_precision',
                        'neg_log_loss', 'neg_brier_score', 'r2', 'neg_mean_squared_error', 'neg_mean_absolute_error',
                         'neg_root_mean_squared_error', 'neg_mean_squared_log_error',
                         'neg_median_absolute_error']}],
                'default': 'accuracy'},
            'best_score': {
                'type': 'number',
                'default': 0.0},
            'max_opt_time': {
                'anyOf': [
                {   'type': 'number',
                    'minimum': 0.0},
                {   'enum': [None]}],
                'default': None},
            'lale_num_grids': {
                'anyOf': [
                {   'description': 'If not set, keep all grids.',
                    'enum': [None]},
                {   'description': 'Fraction of grids to keep.',
                    'type': 'number',
                    'minimum': 0.0,
                    'exclusiveMinimum': True,
                    'maximum': 1.0,
                    'exclusiveMaximum': True},
                {   'description': 'Number of grids to keep.',
                    'type': 'integer',
                    'minimum': 1}],
                'default': None}
                }}]}

_input_fit_schema = {
    'type': 'object',
    'required': ['X', 'y'],
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'anyOf': [
                {   'type': 'array', 'items': {'type': ['number', 'string']}},
                {   'type': 'string'}]}},
        'y': {
            'type': 'array', 'items': {'type': 'number'}}}}

_input_predict_schema = {
    'type': 'object',
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'anyOf': [
                {   'type': 'array', 'items': {'type': ['number', 'string']}},
                {   'type': 'string'}]}}}}

_output_predict_schema = {
    'type': 'array', 'items': {'type': 'number'}}

_combined_schemas = {
    'documentation_url': 'https://lale.readthedocs.io/en/latest/modules/lale.lib.lale.hyperopt_cv.html',
    'description': 'SMAC, the optimizer used inside auto-weka and auto-sklearn.',
    'type': 'object',
    'tags': {
        'pre': [],
        'op': ['estimator'],
        'post': []},
    'properties': {
        'hyperparams': _hyperparams_schema,
        'input_fit': _input_fit_schema,
        'input_predict': _input_predict_schema,
        'output_predict': _output_predict_schema}}

lale.docstrings.set_docstrings(SMACImpl, _combined_schemas)

SMAC = lale.operators.make_operator(SMACImpl, _combined_schemas)



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys

__version__ = '0.3.15'

try:
    # This variable is injected in the __builtins__ by the build
    # process. It is used to not try to import rest of the lale packages when 
    # it is being installed.
    __LALE_SETUP__ # type: ignore
except NameError:
    __LALE_SETUP__ = False

if __LALE_SETUP__:
    sys.stderr.write('Partial import of lale during the build process.\n')
    # We are not importing the rest of lale during the build
    # process.
else:
    #all other code will go here.
    from .operator_wrapper import wrap_imported_operators



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Scikit-learn compatible wrappers for a subset of the operators from imbalanced-learn_ along with schemas to enable hyperparameter tuning.

.. _imbalanced-learn: https://imbalanced-learn.readthedocs.io/en/stable/index.html

Operators:
==========
* `CondensedNearestNeighbour`_
* `EditedNearestNeighbours`_
* `RepeatedEditedNearestNeighbours`_
* `AllKNN`_
* `InstanceHardnessThreshold`_
* `ADASYN`_
* `BorderlineSMOTE`_
* `RandomOverSampler`_
* `SMOTE`_
* `SVMSMOTE`_
* `SMOTEENN`_

.. _`CondensedNearestNeighbour`: lale.lib.imblearn.condensed_nearest_neighbour.html
.. _`EditedNearestNeighbours`: lale.lib.imblearn.edited_nearest_neighbours.html
.. _`RepeatedEditedNearestNeighbours`: lale.lib.imblearn.repeated_edited_nearest_neighbours.html
.. _`AllKNN`: lale.lib.imblearn.all_knn.html
.. _`InstanceHardnessThreshold`: lale.lib.imblearn.instance_hardness_threshold.html
.. _`ADASYN`: lale.lib.imblearn.adasyn.html
.. _`BorderlineSMOTE`: lale.lib.imblearn.borderline_smote.html
.. _`RandomOverSampler`: lale.lib.imblearn.random_over_sampler.html
.. _`SMOTE`: lale.lib.imblearn.smote.html
.. _`SVMSMOTE`: lale.lib.imblearn.svm_smote.html
.. _`SMOTEENN`: lale.lib.imblearn.smoteenn.html

"""

from .smote import SMOTE
from .smoteenn import SMOTEENN
from .condensed_nearest_neighbour import CondensedNearestNeighbour
from .edited_nearest_neighbours import EditedNearestNeighbours
from .repeated_edited_nearest_neighbours import RepeatedEditedNearestNeighbours
from .all_knn import AllKNN
from .instance_hardness_threshold import InstanceHardnessThreshold
from .adasyn import ADASYN
from .borderline_smote import BorderlineSMOTE
from .svm_smote import SVMSMOTE
from .random_over_sampler import RandomOverSampler



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import lale.operators
import lale.docstrings
import sklearn.preprocessing

class OneHotEncoderImpl():
    def __init__(self, **hyperparams):
        self._hyperparams = hyperparams
        self._wrapped_model = sklearn.preprocessing.OneHotEncoder(**self._hyperparams)

    def fit(self, X, y=None):
        self._wrapped_model.fit(X, y)
        return self

    def transform(self, X):
        return self._wrapped_model.transform(X)


_hyperparams_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Hyperparameter schema for the OneHotEncoder model from scikit-learn.',
    'allOf': [
    {   'description': 'This first object lists all constructor arguments with their types, but omits constraints for conditional hyperparameters.',
        'type': 'object',
        'additionalProperties': False,
        'required': ['categories', 'sparse', 'dtype', 'handle_unknown'],
        'relevantToOptimizer': [],
        'properties': {
            'categories': {
                'anyOf': [
                {   'description':
                        'Determine categories automatically from training data.',
                    'enum': ['auto', None]},
                {   'description': 'The ith list element holds the categories expected in the ith column.',
                    'type': 'array',
                    'items': {
                        'anyOf': [
                        {   'type': 'array',
                            'items': {
                                'type': 'string'},
                        }, {
                            'type': 'array',
                            'items': {
                                'type': 'number'},
                            'description': 'Should be sorted.'}]}}],
                'default': 'auto'},
            'sparse': {
                'description':
                    'Will return sparse matrix if set true, else array.',
                'type': 'boolean',
                'default': True},
            'dtype': {
                'description': 'Desired dtype of output, must be number. See https://docs.scipy.org/doc/numpy-1.14.0/reference/arrays.scalars.html#arrays-scalars-built-in',
                'laleType':'Any',
                'default': 'float64'},
            'handle_unknown': {
                'description': 'Whether to raise an error or ignore if an unknown categorical feature is present during transform.',
                'enum': ['error', 'ignore'],
                'default': 'error'},
        }}]}

_input_fit_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Input data schema for training the OneHotEncoder model from scikit-learn.',
    'type': 'object',
    'required': ['X'],
    'additionalProperties': False,
    'properties': {
        'X': {
            'description': 'Features; the outer array is over samples.',
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'anyOf':[{'type': 'number'}, {'type':'string'}]},
            }},
        'y': {
            'description': 'Target class labels; the array is over samples.'}}}

_input_transform_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Input data schema for predictions using the OneHotEncoder model from scikit-learn.',
    'type': 'object',
    'required': ['X'],
    'additionalProperties': False,
    'properties': {
        'X': {
            'description': 'Features; the outer array is over samples.',
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'anyOf':[{'type': 'number'}, {'type':'string'}]}}}}}

_output_transform_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Output data schema for predictions (projected data) using the OneHotEncoder model from scikit-learn. See the official documentation for details: https://scikit-learn.org/0.20/modules/generated/sklearn.preprocessing.OneHotEncoder.html\n',
    'type': 'array',
    'items': {
        'type': 'array',
        'items': {
            'type': 'number'}}}

_combined_schemas = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': """`One-hot encoder`_ transformer from scikit-learn that encodes categorical features as numbers.

.. _`One-hot encoder`: https://scikit-learn.org/0.20/modules/generated/sklearn.preprocessing.OneHotEncoder.html#sklearn-preprocessing-onehotencoder
""",
    'documentation_url': 'https://lale.readthedocs.io/en/latest/modules/lale.lib.sklearn.one_hot_encoder.html',
    'type': 'object',
    'tags': {
        'pre': ['categoricals'],
        'op': ['transformer'],
        'post': []},
    'properties': {
        'hyperparams': _hyperparams_schema,
        'input_fit': _input_fit_schema,
        'input_transform': _input_transform_schema,
        'output_transform': _output_transform_schema}}

lale.docstrings.set_docstrings(OneHotEncoderImpl, _combined_schemas)

OneHotEncoder = lale.operators.make_operator(OneHotEncoderImpl, _combined_schemas)



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import lale.docstrings
import lale.helpers
import lale.operators
import numpy as np

class BatchingImpl():
  def __init__(self, operator = None, batch_size = 32, shuffle = True, num_workers = 0, inmemory=False):    
    self.operator = operator
    self.batch_size = batch_size
    self.shuffle = shuffle
    self.num_workers = num_workers
    self.inmemory = inmemory

  def fit(self, X, y = None):
    if self.operator is None:
      raise ValueError("The pipeline object can't be None at the time of fit.")
    data_loader = lale.helpers.create_data_loader(X = X, y = y, batch_size = self.batch_size)
    classes = np.unique(y)
    self.operator = self.operator.fit_with_batches(data_loader, y = classes, serialize = self.inmemory)
    return self

  def transform(self, X, y = None):
    data_loader = lale.helpers.create_data_loader(X = X, y = y, batch_size = self.batch_size)
    transformed_data = self.operator.transform_with_batches(data_loader, serialize = self.inmemory)
    return transformed_data

  def predict(self, X, y = None):
    return self.transform(X, y)

_input_fit_schema = {
  '$schema': 'http://json-schema.org/draft-04/schema#',
  'description': 'Input data schema for fit.',
  'type': 'object',
  'required': ['X', 'y'],
  'additionalProperties': False,
  'properties': {
    'X': {
      'description': 'Features; the outer array is over samples.',
      'anyOf': [
        { 'type': 'array',
          'items': {'type': 'number'}},
        { 'type': 'array',
          'items': {
            'type': 'array',
            'items': {'type': 'number'}}}]},
    'y': {
      'type': 'array',
      'items': {'anyOf':[{'type': 'integer'}, {'type':'number'}]}
    }
  }
}

_input_predict_transform_schema = { #TODO: separate predict vs. transform
  '$schema': 'http://json-schema.org/draft-04/schema#',
  'description': 'Input data schema for predictions.',
  'type': 'object',
  'required': ['X'],
  'additionalProperties': False,
  'properties': {
    'X': {
      'description': 'Features; the outer array is over samples.',
      'anyOf': [
        { 'type': 'array',
          'items': {'type': 'number'}},
        { 'type': 'array',
          'items': {
            'type': 'array',
            'items': {'type': 'number'}}}]},
    'y': {
      'type': 'array',
      'items': {'anyOf':[{'type': 'integer'}, {'type':'number'}]}
    }
  }
}

_output_schema = { #TODO: separate predict vs. transform
  '$schema': 'http://json-schema.org/draft-04/schema#',
  'description': 'Output data schema for transformed data.',
  'laleType': 'Any'}
  

_hyperparams_schema = {
  '$schema': 'http://json-schema.org/draft-04/schema#',
  'description': 'Hyperparameter schema.',
  'allOf': [
    { 'description':
        'This first sub-object lists all constructor arguments with their '
        'types, one at a time, omitting cross-argument constraints.',
      'type': 'object',
      'additionalProperties': False,
      'relevantToOptimizer': ['batch_size'],
      'properties': {
        'operator':{
          'description':'A lale pipeline object to be used inside of batching',
          'laleType': 'operator'
        },
        'batch_size':{
          'description': 'Batch size used for transform.',
          'type': 'integer',
          'default': 64,
          'minimum': 1,
          'distribution': 'uniform',
          'minimumForOptimizer': 32,
          'maximumForOptimizer': 128},
        'shuffle':{
          'type':'boolean',
          'default': False,
          'description': 'Shuffle dataset before batching or not.'
          },
        'num_workers':{
          'type':'integer',
          'default':0,
          'description': 'Number of workers for pytorch dataloader.'
          },
        'inmemory':{
          'type':'boolean',
          'default': False,
          'description': 'Whether all the computations are done in memory or intermediate outputs are serialized.'
          }
          }}]}

_combined_schemas = {
  '$schema': 'http://json-schema.org/draft-04/schema#',
  'description': """Batching trains the given pipeline using batches.
The batch_size is used across all steps of the pipeline, serializing
the intermediate outputs if specified.""",
  'type': 'object',
  'tags': {
    'pre': [],
    'op': ['estimator', 'transformer'],
    'post': []},
  'properties': {
    'hyperparams': _hyperparams_schema,
    'input_fit': _input_fit_schema,
    'input_predict': _input_predict_transform_schema,
    'output_predict': _output_schema,
    'input_transform': _input_predict_transform_schema,
    'output_transform': _output_schema}}

lale.docstrings.set_docstrings(BatchingImpl, _combined_schemas)

Batching = lale.operators.make_operator(BatchingImpl, _combined_schemas)



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib
import inspect
import jsonschema
import keyword
import lale.operators
import re
from typing import Any, Dict, Optional, Tuple, cast
import logging

logger = logging.getLogger(__name__)

JSON_TYPE = Dict[str, Any]

SCHEMA = {
  '$schema': 'http://json-schema.org/draft-04/schema#',
  'definitions': {
    'operator': {
      'anyOf': [
        {'$ref': '#/definitions/planned_individual_op'},
        {'$ref': '#/definitions/trainable_individual_op'},
        {'$ref': '#/definitions/trained_individual_op'},
        {'$ref': '#/definitions/planned_pipeline'},
        {'$ref': '#/definitions/trainable_pipeline'},
        {'$ref': '#/definitions/trained_pipeline'},
        {'$ref': '#/definitions/operator_choice'}]},
    'individual_op': {
      'type': 'object',
      'required': ['class', 'state', 'operator'],
      'properties': {
        'class': {
          'type': 'string',
          'pattern': '^([A-Za-z_][A-Za-z_0-9]*[.])*[A-Za-z_][A-Za-z_0-9]*$'},
        'state': {
          'enum': ['metamodel', 'planned', 'trainable', 'trained'] },
        'operator': {
          'type': 'string',
          'pattern': '^[A-Za-z_][A-Za-z_0-9]*$'},
        'label': {
          'type': 'string',
          'pattern': '^[A-Za-z_][A-Za-z_0-9]*$'},
        'documentation_url': {
          'type': 'string'},
        'hyperparams': {
          'anyOf': [
            { 'enum': [None]},
            { 'type': 'object',
              'patternProperties': {'^[A-Za-z_][A-Za-z_0-9]*$': {}}}]},
        'steps': {
          'description': 'Nested operators in higher-order individual op.',
          'type': 'object',
          'patternProperties': {
            '^[a-z][a-z_0-9]*$': {'$ref': '#/definitions/operator'}}},
        'is_frozen_trainable': {
          'type': 'boolean'},
        'is_frozen_trained': {
          'type': 'boolean'},
        'coefs': {
          'enum': [None, 'coefs_not_available']}}},
    'planned_individual_op': {
      'allOf': [
        { '$ref': '#/definitions/individual_op'},
        { 'type': 'object',
          'properties': { 'state': { 'enum': ['planned']}}}]},
    'trainable_individual_op': {
      'allOf': [
        { '$ref': '#/definitions/individual_op'},
        { 'type': 'object',
          'required': ['hyperparams', 'is_frozen_trainable'],
          'properties': { 'state': { 'enum': ['trainable']}}}]},
    'trained_individual_op': {
      'allOf': [
        { '$ref': '#/definitions/individual_op'},
        { 'type': 'object',
          'required': ['hyperparams', 'coefs', 'is_frozen_trained'],
          'properties': { 'state': { 'enum': ['trained']}}}]},
    'pipeline': {
      'type': 'object',
      'required': ['class', 'state', 'edges', 'steps'],
      'properties': {
        'class': {
          'enum': [
            'lale.operators.PlannedPipeline',
            'lale.operators.TrainablePipeline',
            'lale.operators.TrainedPipeline']},
        'state': {
          'enum': ['planned', 'trainable', 'trained']},
        'edges': {
          'type': 'array',
          'items': {
            'type': 'array',
            'minItems': 2, 'maxItems': 2,
            'items': {
              'type': 'string',
              'pattern': '^[a-z][a-z_0-9]*$'}}},
        'steps': {
          'type': 'object',
          'patternProperties': {
            '^[a-z][a-z_0-9]*$': {'$ref': '#/definitions/operator'}}}}},
    'planned_pipeline': {
      'allOf': [
        { '$ref': '#/definitions/pipeline'},
        { 'type': 'object',
          'properties': {
            'state': { 'enum': ['planned']},
            'class': { 'enum': ['lale.operators.PlannedPipeline']}}}]},
    'trainable_pipeline': {
      'allOf': [
        { '$ref': '#/definitions/pipeline'},
        { 'type': 'object',
          'properties': {
            'state': { 'enum': ['trainable']},
            'class': { 'enum': ['lale.operators.TrainablePipeline']},
            'steps': {
              'type': 'object',
              'patternProperties': {
                '^[a-z][a-z_0-9]*$': {
                  'type': 'object',
                  'properties': {
                    'state': { 'enum': ['trainable', 'trained']}}}}}}}]},
    'trained_pipeline': {
      'allOf': [
        { '$ref': '#/definitions/pipeline'},
        { 'type': 'object',
          'properties': {
            'state': { 'enum': ['trained']},
            'class': { 'enum': ['lale.operators.TrainedPipeline']},
            'steps': {
              'type': 'object',
              'patternProperties': {
                '^[a-z][a-z_0-9]*$': {
                  'type': 'object',
                  'properties': {
                    'state': { 'enum': ['trained']}}}}}}}]},
    'operator_choice': {
      'type': 'object',
      'required': ['class', 'state', 'operator', 'steps'],
      'properties': {
        'class': {
          'enum': ['lale.operators.OperatorChoice']},
        'state': {
          'enum': ['planned']},
        'operator': {
          'type': 'string'},
        'steps': {
          'type': 'object',
          'patternProperties': {
            '^[a-z][a-z_0-9]*$': {'$ref': '#/definitions/operator'}}}}}},
  '$ref': '#/definitions/operator'}

if __name__ == "__main__":
    import lale.type_checking
    lale.type_checking.validate_is_schema(SCHEMA)

def json_op_kind(jsn: JSON_TYPE) -> str:
    if jsn['class'] == 'lale.operators.OperatorChoice':
        return 'OperatorChoice'
    if jsn['class'] in ['lale.operators.PlannedPipeline',
                        'lale.operators.TrainablePipeline',
                        'lale.operators.TrainedPipeline']:
        return 'Pipeline'
    return 'IndividualOp'

def _get_state(op: 'lale.operators.Operator') -> str:
    if isinstance(op, lale.operators.TrainedOperator):
        return 'trained'
    if isinstance(op, lale.operators.TrainableOperator):
        return 'trainable'
    if isinstance(op, lale.operators.PlannedOperator) or isinstance(op, lale.operators.OperatorChoice):
        return 'planned'
    if isinstance(op, lale.operators.Operator):
        return 'metamodel'
    raise TypeError(f'Expected lale.operators.Operator, got {type(op)}.')

def _get_cls2label(call_depth: int) -> Dict[str, str]:
    frame = inspect.stack()[call_depth][0]
    cls2label: Dict[str, str] = {}
    cls2state: Dict[str, str] = {}
    all_items = [*frame.f_locals.items(), *frame.f_globals.items()]
    for label, op in all_items:
        if isinstance(op, lale.operators.IndividualOp):
            state = _get_state(op)
            cls = op.class_name()
            if cls in cls2state:
                insert = (
                    (cls2state[cls] == 'trainable' and state == 'planned') or
                    (cls2state[cls] == 'trained' and
                     state in ['trainable', 'planned']))
            else:
                insert = True
            if insert:
                cls2label[cls] = label
                cls2state[cls] = state
    return cls2label

def _camelCase_to_snake(name):
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()

class _GenSym:
    def __init__(self, op: 'lale.operators.Operator', cls2label: Dict[str, str]):
        label2count: Dict[str, int] = {}
        def populate_label2count(op: 'lale.operators.Operator'):
            if isinstance(op, lale.operators.IndividualOp):
                label = cls2label.get(op.class_name(), op.name())
            elif isinstance(op, lale.operators.BasePipeline):
                for s in op.steps():
                    populate_label2count(s)
                label = 'pipeline'
            elif isinstance(op, lale.operators.OperatorChoice):
                for s in op.steps():
                    populate_label2count(s)
                label = 'choice'
            label2count[label] = label2count.get(label, 0) + 1
        populate_label2count(op)
        non_unique_labels = {l for l, c in label2count.items() if c > 1}
        snakes = {_camelCase_to_snake(l) for l in non_unique_labels}
        self._names = ({'lale', 'make_pipeline', 'make_union', 'make_choice'}
                       | set(keyword.kwlist) | non_unique_labels | snakes)

    def __call__(self, prefix: str) -> str:
        if prefix in self._names:
            suffix = 0
            while f'{prefix}_{suffix}' in self._names:
                suffix += 1
            result = f'{prefix}_{suffix}'
        else:
            result = prefix
        self._names |= {result}
        return result

def _hps_to_json_rec(hps, cls2label: Dict[str, str], gensym: _GenSym, steps) -> Any:
    if isinstance(hps, lale.operators.Operator):
        step_uid, step_jsn = _op_to_json_rec(hps, cls2label, gensym)
        steps[step_uid] = step_jsn
        return {'$ref': f'../steps/{step_uid}'}
    elif isinstance(hps, dict):
        return {hp_name: _hps_to_json_rec(hp_val, cls2label, gensym, steps)
               for hp_name, hp_val in hps.items()}
    elif isinstance(hps, tuple):
        return tuple([_hps_to_json_rec(hp_val, cls2label, gensym, steps)
                      for hp_val in hps])
    elif isinstance(hps, list):
        return [_hps_to_json_rec(hp_val, cls2label, gensym, steps)
                for hp_val in hps]
    else:
        return hps

def _op_to_json_rec(op: 'lale.operators.Operator', cls2label: Dict[str, str], gensym: _GenSym) -> Tuple[str, JSON_TYPE]:
    jsn: JSON_TYPE = {}
    jsn['class'] = op.class_name()
    jsn['state'] = _get_state(op)
    if isinstance(op, lale.operators.IndividualOp):
        jsn['operator'] = op.name()
        jsn['label'] = cls2label.get(op.class_name(), op.name())
        uid = gensym(_camelCase_to_snake(jsn['label']))
        documentation_url = op.documentation_url()
        if documentation_url is not None:
            jsn['documentation_url'] = documentation_url
        if isinstance(op, lale.operators.TrainableIndividualOp):
            if op.hyperparams() is None:
                jsn['hyperparams'] = None
            else:
                steps: Dict[str, JSON_TYPE] = {}
                jsn['hyperparams'] = _hps_to_json_rec(
                    op.hyperparams(), cls2label, gensym, steps)
                if len(steps) > 0:
                    jsn['steps'] = steps
            jsn['is_frozen_trainable'] = op.is_frozen_trainable()
        if isinstance(op, lale.operators.TrainedIndividualOp):
            if hasattr(op._impl, 'fit'):
                jsn['coefs'] = 'coefs_not_available'
            else:
                jsn['coefs'] = None
            jsn['is_frozen_trained'] = op.is_frozen_trained()
    elif isinstance(op, lale.operators.BasePipeline):
        uid = gensym('pipeline')
        child2uid: Dict[lale.operators.Operator, str] = {}
        child2jsn: Dict[lale.operators.Operator, JSON_TYPE] = {}
        for idx, child in enumerate(op.steps()):
            child_uid, child_jsn = _op_to_json_rec(child, cls2label, gensym)
            child2uid[child] = child_uid
            child2jsn[child] = child_jsn
        jsn['edges'] = [[child2uid[x], child2uid[y]] for x, y in op.edges()]
        jsn['steps'] = {child2uid[z]: child2jsn[z] for z in op.steps()}
    elif isinstance(op, lale.operators.OperatorChoice):
        jsn['operator'] = 'OperatorChoice'
        uid = gensym('choice')
        jsn['state'] = 'planned'
        jsn['steps'] = {}
        for step in op.steps():
            child_uid, child_jsn = _op_to_json_rec(step, cls2label, gensym)
            jsn['steps'][child_uid] = child_jsn
    return uid, jsn

def to_json(op: 'lale.operators.Operator', call_depth:int=1) -> JSON_TYPE:
    cls2label = _get_cls2label(call_depth + 1)
    gensym = _GenSym(op, cls2label)
    uid, jsn = _op_to_json_rec(op, cls2label, gensym)
    jsonschema.validate(jsn, SCHEMA, jsonschema.Draft4Validator)
    return jsn

def _get_lib_schema(impl) -> Optional[JSON_TYPE]:
    if impl.__module__.startswith('lale.lib'):
        m = importlib.import_module(impl.__module__)
        return getattr(m, '_combined_schemas')
    module_name = impl.__module__.split('.')[0]
    class_name = _camelCase_to_snake(impl.__class__.__name__)
    try:
        lib_name = '.'.join(['lale.lib', module_name, class_name])
        m = importlib.import_module(lib_name)
        return getattr(m, '_combined_schemas')
    except (ModuleNotFoundError, AttributeError):
        try:
            lib_name = '.'.join(['lale.lib.autogen', class_name])
            m = importlib.import_module(lib_name)
            return getattr(m, '_combined_schemas')
        except:
            return None

def _hps_from_json_rec(jsn: JSON_TYPE, steps: JSON_TYPE) -> Any:
    if isinstance(jsn, dict):
        if '$ref' in jsn:
            step_uid = jsn['$ref'].split('/')[-1]
            step_jsn = steps[step_uid]
            return _op_from_json_rec(step_jsn)
        else:
            return {k: _hps_from_json_rec(v, steps) for k, v in jsn.items()}
    elif isinstance(jsn, tuple):
        return tuple([_hps_from_json_rec(v, steps) for v in jsn])
    elif isinstance(jsn, list):
        return [_hps_from_json_rec(v, steps) for v in jsn]
    else:
        return jsn

def _op_from_json_rec(jsn: JSON_TYPE) -> 'lale.operators.Operator':
    kind = json_op_kind(jsn)
    if kind == 'Pipeline':
        steps_dict = {
            uid: _op_from_json_rec(jsn['steps'][uid]) for uid in jsn['steps']}
        steps = [steps_dict[i] for i in steps_dict]
        edges = [(steps_dict[x], steps_dict[y]) for (x,y) in jsn['edges']]
        return lale.operators.get_pipeline_of_applicable_type(steps, edges)
    elif kind == 'OperatorChoice':
        steps = [_op_from_json_rec(s) for s in jsn['steps'].values()]
        name = jsn['operator']
        return lale.operators.OperatorChoice(steps, name)
    else:
        assert kind == 'IndividualOp'
        full_class_name = jsn['class']
        last_period = full_class_name.rfind('.')
        module = importlib.import_module(full_class_name[:last_period])
        impl = getattr(module, full_class_name[last_period+1:])
        schemas = _get_lib_schema(impl)
        name = jsn['operator']
        result = lale.operators.make_operator(impl, schemas, name)
        if jsn['state'] in ['trainable', 'trained']:
            if _get_state(result) == 'planned':
                hps = jsn['hyperparams']
                if hps is None:
                    result = result()
                else:
                    hps = _hps_from_json_rec(hps, jsn.get('steps', {}))
                    result = result(**hps)
            trnbl = cast(lale.operators.TrainableIndividualOp, result)
            if jsn['is_frozen_trainable'] and not trnbl.is_frozen_trainable():
                trnbl = trnbl.freeze_trainable()
            assert jsn['is_frozen_trainable'] == trnbl.is_frozen_trainable()
            result = trnbl
        if jsn['state'] == 'trained':
            if jsn['coefs'] == 'coefs_not_available':
                logger.warning(f'Since the JSON representation of trained operator {name} lacks coefficients, from_json returns a trainable operator instead.')
            else:
                assert jsn['coefs'] is None, jsn['coefs']
        assert _get_state(result) == jsn['state'] or \
            jsn['state']=='trained' and jsn['coefs']=='coefs_not_available'
        assert result.documentation_url() == jsn['documentation_url']
        return result
    assert False, f'unexpected JSON {jsn}'

def from_json(jsn: JSON_TYPE) -> 'lale.operators.Operator':
    jsonschema.validate(jsn, SCHEMA, jsonschema.Draft4Validator)
    result = _op_from_json_rec(jsn)
    return result



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from imblearn.over_sampling import SMOTE as OrigModel
import lale.operators

class BaseResamplerImpl():

    def __init__(self, operator = None, resampler = None):
        self.operator = operator
        self.resampler = resampler

    def fit(self, X, y=None):
        X, y = self.resampler.fit_resample(X, y)
        self.trained_operator = self.operator.fit(X, y)
        return self

    def transform(self, X, y=None):
        return self.trained_operator.transform(X, y)

    def predict(self, X):
        return self.trained_operator.predict(X)

    def predict_proba(self, X):
        return self.trained_operator.predict_proba(X)

    def decision_function(self, X):
        return self.trained_operator.decision_function(X)

_input_fit_schema = {
  '$schema': 'http://json-schema.org/draft-04/schema#',
  'type': 'object',
  'required': ['X', 'y'],
  'additionalProperties': False,
  'properties': {
    'X': {
      'description': 'Features; the outer array is over samples.',
      'type': 'array',
      'items': {'type': 'array', 'items': {'type': 'number'}}},
    'y': {
      'description': 'Target class labels; the array is over samples.',
        'anyOf': [
            {'type': 'array', 'items': {'type': 'number'}},
            {'type': 'array', 'items': {'type': 'string'}}]}}}

_input_transform_schema = {
  '$schema': 'http://json-schema.org/draft-04/schema#',
  'type': 'object',
  'required': ['X', 'y'],
  'additionalProperties': False,
  'properties': {
    'X': {
      'description': 'Features; the outer array is over samples.',
      'type': 'array',
      'items': {'type': 'array', 'items': {'type': 'number'}}},
    'y': {
      'description': 'Target class labels; the array is over samples.',
      'anyOf': [
          {'type': 'array', 'items': {'type': 'number'}},
          {'type': 'array', 'items': {'type': 'string'}},
          {'enum': [None]}]}}}

_output_transform_schema = {
  '$schema': 'http://json-schema.org/draft-04/schema#',
  'description': 'Output data schema for transformed data.',
  'laleType': 'Any'}

_input_predict_schema = {
  '$schema': 'http://json-schema.org/draft-04/schema#',
  'type': 'object',
  'required': ['X'],
  'additionalProperties': False,
  'properties': {
    'X': {
      'description': 'Features; the outer array is over samples.',
      'type': 'array',
      'items': {'type': 'array', 'items': {'type': 'number'}}}
}}

_output_predict_schema={
  '$schema': 'http://json-schema.org/draft-04/schema#',
  'description': 'Output data schema for predictions.',
  'laleType': 'Any'}

_input_predict_proba_schema = {
  '$schema': 'http://json-schema.org/draft-04/schema#',
  'type': 'object',
  'required': ['X'],
  'additionalProperties': False,
  'properties': {
    'X': {
      'description': 'Features; the outer array is over samples.',
      'type': 'array',
      'items': {'type': 'array', 'items': {'type': 'number'}}}
}}

_output_predict_proba_schema={
  '$schema': 'http://json-schema.org/draft-04/schema#',
  'description': 'Output data schema for predictions.',
  'laleType': 'Any'}

_input_decision_function_schema = {
  '$schema': 'http://json-schema.org/draft-04/schema#',
  'type': 'object',
  'required': ['X'],
  'additionalProperties': False,
  'properties': {
    'X': {
      'description': 'Features; the outer array is over samples.',
      'type': 'array',
      'items': {'type': 'array', 'items': {'type': 'number'}}}
}}

_output_decision_function_schema={
  '$schema': 'http://json-schema.org/draft-04/schema#',
  'description': 'Output data schema for predictions.',
  'laleType': 'Any'}




from sklearn.decomposition.factor_analysis import FactorAnalysis as Op
import lale.helpers
import lale.operators
import lale.docstrings
from numpy import nan, inf

class FactorAnalysisImpl():

    def __init__(self, n_components=None, tol=0.01, copy=True, max_iter=1000, noise_variance_init=None, svd_method='randomized', iterated_power=3, random_state=0):
        self._hyperparams = {
            'n_components': n_components,
            'tol': tol,
            'copy': copy,
            'max_iter': max_iter,
            'noise_variance_init': noise_variance_init,
            'svd_method': svd_method,
            'iterated_power': iterated_power,
            'random_state': random_state}
        self._wrapped_model = Op(**self._hyperparams)

    def fit(self, X, y=None):
        if (y is not None):
            self._wrapped_model.fit(X, y)
        else:
            self._wrapped_model.fit(X)
        return self

    def transform(self, X):
        return self._wrapped_model.transform(X)
_hyperparams_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'inherited docstring for FactorAnalysis    Factor Analysis (FA)',
    'allOf': [{
        'type': 'object',
        'required': ['n_components', 'tol', 'copy', 'max_iter', 'noise_variance_init', 'svd_method', 'iterated_power', 'random_state'],
        'relevantToOptimizer': ['n_components', 'tol', 'copy', 'max_iter', 'svd_method', 'iterated_power'],
        'additionalProperties': False,
        'properties': {
            'n_components': {
                'enum': ['int', None],
                'default': None,
                'description': 'Dimensionality of latent space, the number of components of ``X`` that are obtained after ``transform``'},
            'tol': {
                'type': 'number',
                'minimumForOptimizer': 1e-08,
                'maximumForOptimizer': 0.01,
                'distribution': 'loguniform',
                'default': 0.01,
                'description': 'Stopping tolerance for EM algorithm.'},
            'copy': {
                'type': 'boolean',
                'default': True,
                'description': 'Whether to make a copy of X'},
            'max_iter': {
                'type': 'integer',
                'minimumForOptimizer': 10,
                'maximumForOptimizer': 1000,
                'distribution': 'uniform',
                'default': 1000,
                'description': 'Maximum number of iterations.'},
            'noise_variance_init': {
                'XXX TODO XXX': 'None | array, shape=(n_features,)',
                'description': 'The initial guess of the noise variance for each feature',
                'enum': [None],
                'default': None},
            'svd_method': {
                'enum': ['lapack', 'randomized'],
                'default': 'randomized',
                'description': 'Which SVD method to use'},
            'iterated_power': {
                'type': 'integer',
                'minimumForOptimizer': 3,
                'maximumForOptimizer': 4,
                'distribution': 'uniform',
                'default': 3,
                'description': 'Number of iterations for the power method'},
            'random_state': {
                'anyOf': [{
                    'type': 'integer'}, {
                    'type': 'object'}, {
                    'enum': [None]}],
                'default': 0,
                'description': 'If int, random_state is the seed used by the random number generator; If RandomState instance, random_state is the random number generator; If None, the random number generator is the RandomState instance used by `np.random`'},
        }}, {
        'XXX TODO XXX': "Parameter: iterated_power > only used if svd_method equals 'randomized'"}, {
        'XXX TODO XXX': "Parameter: random_state > only used when svd_method equals 'randomized'"}],
}
_input_fit_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Fit the FactorAnalysis model to X using EM',
    'type': 'object',
    'required': ['X', 'y'],
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            },
            'description': 'Training data.'},
        'y': {
            
        }},
}
_input_transform_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Apply dimensionality reduction to X using the model.',
    'type': 'object',
    'required': ['X'],
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            },
            'description': 'Training data.'},
    },
}
_output_transform_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'The latent variables of X.',
    'type': 'array',
    'items': {
        'type': 'array',
        'items': {
            'type': 'number'},
    },
}
_combined_schemas = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Combined schema for expected data and hyperparameters.',
    'documentation_url': 'https://scikit-learn.org/0.20/modules/generated/sklearn.decomposition.FactorAnalysis#sklearn-decomposition-factoranalysis',
    'type': 'object',
    'tags': {
        'pre': [],
        'op': ['transformer'],
        'post': []},
    'properties': {
        'hyperparams': _hyperparams_schema,
        'input_fit': _input_fit_schema,
        'input_transform': _input_transform_schema,
        'output_transform': _output_transform_schema},
}
lale.docstrings.set_docstrings(FactorAnalysisImpl, _combined_schemas)
FactorAnalysis = lale.operators.make_operator(FactorAnalysisImpl, _combined_schemas)





from sklearn.decomposition.dict_learning import DictionaryLearning as Op
import lale.helpers
import lale.operators
import lale.docstrings
from numpy import nan, inf

class DictionaryLearningImpl():

    def __init__(self, n_components=None, alpha=1, max_iter=1000, tol=1e-08, fit_algorithm='lars', transform_algorithm='omp', transform_n_nonzero_coefs=None, transform_alpha=None, n_jobs=None, code_init=None, dict_init=None, verbose=False, split_sign=False, random_state=None, positive_code=False, positive_dict=False):
        self._hyperparams = {
            'n_components': n_components,
            'alpha': alpha,
            'max_iter': max_iter,
            'tol': tol,
            'fit_algorithm': fit_algorithm,
            'transform_algorithm': transform_algorithm,
            'transform_n_nonzero_coefs': transform_n_nonzero_coefs,
            'transform_alpha': transform_alpha,
            'n_jobs': n_jobs,
            'code_init': code_init,
            'dict_init': dict_init,
            'verbose': verbose,
            'split_sign': split_sign,
            'random_state': random_state,
            'positive_code': positive_code,
            'positive_dict': positive_dict}
        self._wrapped_model = Op(**self._hyperparams)

    def fit(self, X, y=None):
        if (y is not None):
            self._wrapped_model.fit(X, y)
        else:
            self._wrapped_model.fit(X)
        return self

    def transform(self, X):
        return self._wrapped_model.transform(X)
_hyperparams_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'inherited docstring for DictionaryLearning    Dictionary learning',
    'allOf': [{
        'type': 'object',
        'required': ['n_components', 'alpha', 'max_iter', 'tol', 'fit_algorithm', 'transform_algorithm', 'transform_n_nonzero_coefs', 'transform_alpha', 'n_jobs', 'code_init', 'dict_init', 'verbose', 'split_sign', 'random_state', 'positive_code', 'positive_dict'],
        'relevantToOptimizer': ['n_components', 'alpha', 'max_iter', 'tol', 'fit_algorithm', 'transform_algorithm'],
        'additionalProperties': False,
        'properties': {
            'n_components': {
                'anyOf': [{
                    'type': 'integer',
                    'minimumForOptimizer': 2,
                    'maximumForOptimizer': 256,
                    'distribution': 'uniform'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'number of dictionary elements to extract'},
            'alpha': {
                'type': 'number',
                'minimumForOptimizer': 1e-10,
                'maximumForOptimizer': 1.0,
                'distribution': 'loguniform',
                'default': 1,
                'description': 'sparsity controlling parameter'},
            'max_iter': {
                'type': 'integer',
                'minimumForOptimizer': 10,
                'maximumForOptimizer': 1000,
                'distribution': 'uniform',
                'default': 1000,
                'description': 'maximum number of iterations to perform'},
            'tol': {
                'type': 'number',
                'minimumForOptimizer': 1e-08,
                'maximumForOptimizer': 0.01,
                'distribution': 'loguniform',
                'default': 1e-08,
                'description': 'tolerance for numerical error'},
            'fit_algorithm': {
                'enum': ['lars', 'cd'],
                'default': 'lars',
                'description': 'lars: uses the least angle regression method to solve the lasso problem (linear_model.lars_path) cd: uses the coordinate descent method to compute the Lasso solution (linear_model.Lasso)'},
            'transform_algorithm': {
                'enum': ['lasso_lars', 'lasso_cd', 'lars', 'omp', 'threshold'],
                'default': 'omp',
                'description': 'Algorithm used to transform the data lars: uses the least angle regression method (linear_model.lars_path) lasso_lars: uses Lars to compute the Lasso solution lasso_cd: uses the coordinate descent method to compute the Lasso solution (linear_model.Lasso)'},
            'transform_n_nonzero_coefs': {
                'XXX TODO XXX': 'int, ``0.1 * n_features`` by default',
                'description': 'Number of nonzero coefficients to target in each column of the solution',
                'enum': [None],
                'default': None},
            'transform_alpha': {
                'anyOf': [{
                    'type': 'number'}, {
                    'enum': [None]}],
                'default': None,
                'description': "If `algorithm='lasso_lars'` or `algorithm='lasso_cd'`, `alpha` is the penalty applied to the L1 norm"},
            'n_jobs': {
                'anyOf': [{
                    'type': 'integer'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'Number of parallel jobs to run'},
            'code_init': {
                'anyOf': [{
                    'type': 'array',
                    'items': {
                        'type': 'array',
                        'items': {
                            'type': 'number'},
                    }}, {
                    'enum': [None]}],
                'default': None,
                'description': 'initial value for the code, for warm restart'},
            'dict_init': {
                'anyOf': [{
                    'type': 'array',
                    'items': {
                        'type': 'array',
                        'items': {
                            'type': 'number'},
                    }}, {
                    'enum': [None]}],
                'default': None,
                'description': 'initial values for the dictionary, for warm restart'},
            'verbose': {
                'type': 'boolean',
                'default': False,
                'description': 'To control the verbosity of the procedure.'},
            'split_sign': {
                'type': 'boolean',
                'default': False,
                'description': 'Whether to split the sparse feature vector into the concatenation of its negative part and its positive part'},
            'random_state': {
                'anyOf': [{
                    'type': 'integer'}, {
                    'type': 'object'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'If int, random_state is the seed used by the random number generator; If RandomState instance, random_state is the random number generator; If None, the random number generator is the RandomState instance used by `np.random`.'},
            'positive_code': {
                'type': 'boolean',
                'default': False,
                'description': 'Whether to enforce positivity when finding the code'},
            'positive_dict': {
                'type': 'boolean',
                'default': False,
                'description': 'Whether to enforce positivity when finding the dictionary  '},
        }}, {
        'XXX TODO XXX': "Parameter: transform_n_nonzero_coefs > only used by algorithm='lars' and algorithm='omp' and is overridden by alpha in the omp case"}],
}
_input_fit_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Fit the model from data in X.',
    'type': 'object',
    'required': ['X', 'y'],
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            },
            'description': 'Training vector, where n_samples in the number of samples and n_features is the number of features.'},
        'y': {
            
        }},
}
_input_transform_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Encode the data as a sparse combination of the dictionary atoms.',
    'type': 'object',
    'required': ['X'],
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            },
            'description': 'Test data to be transformed, must have the same number of features as the data used to train the model.'},
    },
}
_output_transform_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Transformed data',
    'type': 'array',
    'items': {
        'type': 'array',
        'items': {
            'type': 'number'},
    },
}
_combined_schemas = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Combined schema for expected data and hyperparameters.',
    'documentation_url': 'https://scikit-learn.org/0.20/modules/generated/sklearn.decomposition.DictionaryLearning#sklearn-decomposition-dictionarylearning',
    'type': 'object',
    'tags': {
        'pre': [],
        'op': ['transformer'],
        'post': []},
    'properties': {
        'hyperparams': _hyperparams_schema,
        'input_fit': _input_fit_schema,
        'input_transform': _input_transform_schema,
        'output_transform': _output_transform_schema},
}
lale.docstrings.set_docstrings(DictionaryLearningImpl, _combined_schemas)
DictionaryLearning = lale.operators.make_operator(DictionaryLearningImpl, _combined_schemas)




# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sklearn.svm.classes
import lale.docstrings
import lale.operators

class LinearSVCImpl():
    def __init__(self, penalty=None, loss=None, dual=True, tol=0.0001, C=1.0, multi_class=None, fit_intercept=True, intercept_scaling=1, class_weight=None, verbose=0, random_state=None, max_iter=1000):
        self._hyperparams = {
            'penalty': penalty,
            'loss': loss,
            'dual': dual,
            'tol': tol,
            'C': C,
            'multi_class': multi_class,
            'fit_intercept': fit_intercept,
            'intercept_scaling': intercept_scaling,
            'class_weight': class_weight,
            'verbose': verbose,
            'random_state': random_state,
            'max_iter': max_iter}
        self._wrapped_model = sklearn.svm.classes.LinearSVC(**self._hyperparams)

    def fit(self, X, y=None, sample_weight=None):
        self._wrapped_model.fit(X, y)
        return self

    def predict(self, X):
        return self._wrapped_model.predict(X)

    def decision_function(self, X):
        return self._wrapped_model.decision_function(X)

_hyperparams_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Linear Support Vector Classification.',
    'allOf': [
    {   'type': 'object',
        'additionalProperties': False,
        'required': [
            'penalty', 'loss', 'dual', 'tol', 'C', 'multi_class',
            'fit_intercept', 'intercept_scaling', 'class_weight',
            'verbose', 'random_state', 'max_iter'],
        'relevantToOptimizer': [
            'penalty', 'loss', 'dual', 'tol', 'C', 'multi_class',
            'fit_intercept'],
        'properties': {
            'penalty': {
                'description': 'Norm used in the penalization.',
                'enum': ['l1', 'l2'],
                'default': 'l2'},
            'loss': {
                'description': 'Loss function.',
                'enum': ['hinge', 'squared_hinge'],
                'default': 'squared_hinge'},
            'dual': {
                'type': 'boolean',
                'default': True,
                'description': 'Select the algorithm to either solve the dual or primal optimization problem.'},
            'tol': {
                'type': 'number',
                'distribution': 'loguniform',
                'minimumForOptimizer': 1e-08,
                'maximumForOptimizer': 0.01,
                'default': 0.0001,
                'description': 'Tolerance for stopping criteria.'},
            'C': {
                'description': 'Penalty parameter C of the error term.',
                'type': 'number',
                'distribution': 'loguniform',
                'minimum': 0.0,
                'exclusiveMinimum': True,
                'default': 1.0,
                'minimumForOptimizer': 0.03125,
                'maximumForOptimizer': 32768},
            'multi_class': {
                'description': 'Determines the multi-class strategy if `y` contains more than two classes.',
                'enum': ['ovr', 'crammer_singer'],
                'default': 'ovr'},
            'fit_intercept': {
                'type': 'boolean',
                'default': True,
                'description': 'Whether to calculate the intercept for this model.'},
            'intercept_scaling': {
                'type': 'number',
                'description':
                    'Append a constant feature with constant value '
                    'intercept_scaling to the instance vector.',
                'minimumForOptimizer': 0.0,
                'maximumForOptimizer': 1.0,
                'default': 1},
            'class_weight': {
                'anyOf': [
                    { 'description': 'By default, all classes have weight 1.',
                      'enum': [None]},
                    { 'description': 'Adjust weights by inverse frequency.',
                      'enum': ['balanced']},
                    { 'description': 'Dictionary mapping class labels to weights.',
                      'type': 'object',
                      'propertyNames': {'pattern': '^.+$', 'type': 'number'},
                      'forOptimizer': False}],
                'default': None},
            'verbose': {
                'type': 'integer',
                'default': 0,
                'description': 'Enable verbose output.'},
            'random_state': {
                'description':
                'Seed of pseudo-random number generator for shuffling data.',
                'anyOf': [
                    { 'description': 'RandomState used by np.random',
                      'enum': [None]},
                    { 'description': 'Explicit seed.',
                      'type': 'integer'}],
                'default': None},
            'max_iter': {
                'type': 'integer',
                'minimumForOptimizer': 10,
                'maximumForOptimizer': 1000,
                'default': 1000,
                'description': 'The maximum number of iterations to be run.'}}},
    {   'description':
            'The combination of penalty=`l1` and loss=`hinge` is not supported',
        'anyOf': [
            { 'type': 'object',
              'properties': {'penalty': {'enum': ['l2']}}},
            { 'type': 'object',
              'properties': {'loss': {'enum': ['squared_hinge']}}}]},
    {   'description': 'The combination of penalty=`l2` and loss=`hinge` '
                       'is not supported when dual=False.',
        'anyOf': [
            { 'type': 'object',
              'properties': {'penalty': {'enum': ['l1']}}},
            { 'type': 'object',
              'properties': {'loss': {'enum': ['squared_hinge']}}},
            { 'type': 'object',
              'properties': {'dual': {'enum': [True]}}}]},
    {   'description': 'The combination of penalty=`l1` and '
                       'loss=`squared_hinge` is not supported when dual=True.',
        'anyOf': [
            { 'type': 'object',
              'properties': {'penalty': {'enum': ['l2']}}},
            { 'type': 'object',
              'properties': {'loss': {'enum': ['hinge']}}},
            { 'type': 'object',
              'properties': {'dual': {'enum': [False]}}}]}
    ]}

_input_fit_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Fit the model according to the given training data.',
    'type': 'object',
    'required': ['X', 'y'],
    'properties': {
        'X': {
            'type': 'array',
            'items': {'type': 'array', 'items': {'type': 'number'}},
            'description': 'Training vector.'},
        'y': {
            'anyOf': [
                {'type': 'array', 'items': {'type': 'number'}},
                {'type': 'array', 'items': {'type': 'string'}}],
            'description': 'Target vector relative to X.'},
        'sample_weight': {
            'anyOf': [
            {   'type': 'array',
                'items': {'type': 'number'}},
            {   'enum': [None]}],
            'default': None,
            'description':
                'Array of weights that are assigned to individual samples.'}}}

_input_predict_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Predict class labels for samples in X.',
    'type': 'object',
    'required': ['X'],
    'properties': {
        'X': {
            'type': 'array',
            'items': {'type': 'array', 'items': {'type': 'number'}},
            'description': 'Samples.'}}}

_output_predict_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Predict class labels for samples in X.',
    'anyOf': [
        {'type': 'array', 'items': {'type': 'number'}},
        {'type': 'array', 'items': {'type': 'string'}}]}

_input_decision_function_schema = {
  'type': 'object',
  'required': ['X'],
  'additionalProperties': False,
  'properties': {
    'X': {
      'description': 'Features; the outer array is over samples.',
      'type': 'array',
      'items': {'type': 'array', 'items': {'type': 'number'}}}}}

_output_decision_function_schema = {
    'description': 'Confidence scores for samples for each class in the model.',
    'anyOf': [
    {   'description': 'In the multi-way case, score per (sample, class) combination.',
        'type': 'array',
        'items': {'type': 'array', 'items': {'type': 'number'}}},
    {   'description': 'In the binary case, score for `self._classes[1]`.',
        'type': 'array',
        'items': {'type': 'number'}}]}

_combined_schemas = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': """`Linear Support Vector Classification`_ from scikit-learn.

.. _`Linear Support Vector Classification`: https://scikit-learn.org/0.20/modules/generated/sklearn.svm.LinearSVC.html#sklearn-svm-linearsvc
""",
  'documentation_url': 'https://lale.readthedocs.io/en/latest/modules/lale.lib.sklearn.linear_svc.html',
    'type': 'object',
    'tags': {
        'pre': [],
        'op': ['estimator', 'classifier'],
        'post': []},
    'properties': {
        'hyperparams': _hyperparams_schema,
        'input_fit': _input_fit_schema,
        'input_predict': _input_predict_schema,
        'output_predict': _output_predict_schema,
        'input_decision_function': _input_decision_function_schema,
        'output_decision_function': _output_decision_function_schema,
}}

lale.docstrings.set_docstrings(LinearSVCImpl, _combined_schemas)

LinearSVC = lale.operators.make_operator(LinearSVCImpl, _combined_schemas)



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import os
import pandas as pd
import tempfile
import urllib.request
import zipfile
import lale.datasets.data_schemas

download_data_dir = os.path.join(os.path.dirname(__file__), 'download_data')
download_data_url = 'http://archive.ics.uci.edu/ml/machine-learning-databases'

def download(dataset_id, zip_name, contents_files):
    zip_url = f'{download_data_url}/{dataset_id}/{zip_name}'
    data_dir = os.path.join(download_data_dir, dataset_id)
    if not os.path.exists(data_dir):
        os.makedirs(data_dir)
    full_file_names = [os.path.join(data_dir, base) for base in contents_files]
    def all_downloaded():
        for full in full_file_names:
            if not os.path.exists(full):
                return False
        return True
    if not all_downloaded():
        with tempfile.NamedTemporaryFile(suffix='.zip') as tmp_zip_file:
            urllib.request.urlretrieve(zip_url, tmp_zip_file.name)
            with zipfile.ZipFile(tmp_zip_file.name) as myzip:
                for i in range(len(contents_files)):
                    full, base = full_file_names[i], contents_files[i]
                    if not os.path.exists(full):
                        myzip.extract(base, data_dir)
    assert all_downloaded
    return full_file_names

def tsv_to_Xy(file_name, target_col, schema_orig):
    data_all = pd.read_csv(file_name, sep='\t')
    row_schema_X = [col_schema for col_schema in schema_orig['items']['items']
                    if col_schema['description'] != target_col]
    columns_X = [col_schema['description'] for col_schema in row_schema_X]
    data_X = data_all.loc[:, columns_X]
    nrows, ncols_X = data_X.shape
    schema_X = {
        **schema_orig,
        'minItems': nrows, 'maxItems': nrows,
        'items': {
            'type': 'array',
            'minItems': ncols_X, 'maxItems': ncols_X,
            'items': row_schema_X}}
    data_X = lale.datasets.data_schemas.add_schema(data_X, schema_X)
    row_schema_y = [col_schema for col_schema in schema_orig['items']['items']
                    if col_schema['description'] == target_col]
    data_y = data_all[target_col]
    schema_y = {
        **schema_orig,
        'minItems': nrows, 'maxItems': nrows,
        'items': row_schema_y[0]}
    data_y = lale.datasets.data_schemas.add_schema(data_y, schema_y)
    return data_X, data_y

def fetch_drugscom():
    files = download('00462', 'drugsCom_raw.zip',
                     ['drugsComTest_raw.tsv', 'drugsComTrain_raw.tsv'])
    target_col = 'rating'
    json_schema = {
        '$schema': 'http://json-schema.org/draft-04/schema#',
        'type': 'array',
        'items': {
            'type': 'array', 'minItems': 6, 'maxItems': 6,
            'items': [
                {'description': 'drugName', 'type': 'string'},
                {'description': 'condition',
                 'anyOf': [{'type': 'string'}, {'enum': [np.NaN]}]},
                {'description': 'review', 'type': 'string'},
                {'description': 'rating',
                 'enum': [1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 10.0]},
                {'description': 'date', 'type': 'string'},
                {'description': 'usefulCount', 'type': 'integer',
                 'minimum': 0}]}}
    test_X, test_y = tsv_to_Xy(files[0], target_col, json_schema)
    train_X, train_y = tsv_to_Xy(files[1], target_col, json_schema)
    return train_X, train_y, test_X, test_y




from sklearn.linear_model.theil_sen import TheilSenRegressor as Op
import lale.helpers
import lale.operators
import lale.docstrings
from numpy import nan, inf

class TheilSenRegressorImpl():

    def __init__(self, fit_intercept=True, copy_X=True, max_subpopulation=10000, n_subsamples=None, max_iter=300, tol=0.001, random_state=None, n_jobs=None, verbose=False):
        self._hyperparams = {
            'fit_intercept': fit_intercept,
            'copy_X': copy_X,
            'max_subpopulation': max_subpopulation,
            'n_subsamples': n_subsamples,
            'max_iter': max_iter,
            'tol': tol,
            'random_state': random_state,
            'n_jobs': n_jobs,
            'verbose': verbose}
        self._wrapped_model = Op(**self._hyperparams)

    def fit(self, X, y=None):
        if (y is not None):
            self._wrapped_model.fit(X, y)
        else:
            self._wrapped_model.fit(X)
        return self

    def predict(self, X):
        return self._wrapped_model.predict(X)
_hyperparams_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'inherited docstring for TheilSenRegressor    Theil-Sen Estimator: robust multivariate regression model.',
    'allOf': [{
        'type': 'object',
        'required': ['fit_intercept', 'copy_X', 'max_subpopulation', 'n_subsamples', 'max_iter', 'tol', 'random_state', 'n_jobs', 'verbose'],
        'relevantToOptimizer': ['fit_intercept', 'copy_X', 'max_subpopulation', 'max_iter', 'tol'],
        'additionalProperties': False,
        'properties': {
            'fit_intercept': {
                'type': 'boolean',
                'default': True,
                'description': 'Whether to calculate the intercept for this model'},
            'copy_X': {
                'type': 'boolean',
                'default': True,
                'description': 'If True, X will be copied; else, it may be overwritten.'},
            'max_subpopulation': {
                'type': 'integer',
                'minimumForOptimizer': 10000,
                'maximumForOptimizer': 10001,
                'distribution': 'uniform',
                'default': 10000,
                'description': "Instead of computing with a set of cardinality 'n choose k', where n is the number of samples and k is the number of subsamples (at least number of features), consider only a stochastic subpopulation of a given maximal size if 'n choose k' is larger than max_subpopulation"},
            'n_subsamples': {
                'anyOf': [{
                    'type': 'integer'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'Number of samples to calculate the parameters'},
            'max_iter': {
                'type': 'integer',
                'minimumForOptimizer': 10,
                'maximumForOptimizer': 1000,
                'distribution': 'uniform',
                'default': 300,
                'description': 'Maximum number of iterations for the calculation of spatial median.'},
            'tol': {
                'type': 'number',
                'minimumForOptimizer': 1e-08,
                'maximumForOptimizer': 0.01,
                'distribution': 'loguniform',
                'default': 0.001,
                'description': 'Tolerance when calculating spatial median.'},
            'random_state': {
                'anyOf': [{
                    'type': 'integer'}, {
                    'type': 'object'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'A random number generator instance to define the state of the random permutations generator'},
            'n_jobs': {
                'anyOf': [{
                    'type': 'integer'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'Number of CPUs to use during the cross validation'},
            'verbose': {
                'type': 'boolean',
                'default': False,
                'description': 'Verbose mode when fitting the model.'},
        }}, {
        'XXX TODO XXX': "Parameter: max_subpopulation > only a stochastic subpopulation of a given maximal size if 'n choose k' is larger than max_subpopulation"}],
}
_input_fit_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Fit linear model.',
    'type': 'object',
    'required': ['X', 'y'],
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            },
            'description': 'Training data'},
        'y': {
            'type': 'array',
            'items': {
                'type': 'number'},
            'description': 'Target values'},
    },
}
_input_predict_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Predict using the linear model',
    'type': 'object',
    'required': ['X'],
    'properties': {
        'X': {
            'anyOf': [{
                'type': 'array',
                'items': {
                    'laleType': 'Any',
                    'XXX TODO XXX': 'item type'},
                'XXX TODO XXX': 'array_like or sparse matrix, shape (n_samples, n_features)'}, {
                'type': 'array',
                'items': {
                    'type': 'array',
                    'items': {
                        'type': 'number'},
                }}],
            'description': 'Samples.'},
    },
}
_output_predict_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Returns predicted values.',
    'type': 'array',
    'items': {
        'type': 'number'},
}
_combined_schemas = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Combined schema for expected data and hyperparameters.',
    'documentation_url': 'https://scikit-learn.org/0.20/modules/generated/sklearn.linear_model.TheilSenRegressor#sklearn-linear_model-theilsenregressor',
    'type': 'object',
    'tags': {
        'pre': [],
        'op': ['estimator', 'regressor'],
        'post': []},
    'properties': {
        'hyperparams': _hyperparams_schema,
        'input_fit': _input_fit_schema,
        'input_predict': _input_predict_schema,
        'output_predict': _output_predict_schema},
}
lale.docstrings.set_docstrings(TheilSenRegressorImpl, _combined_schemas)
TheilSenRegressor = lale.operators.make_operator(TheilSenRegressorImpl, _combined_schemas)




# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import lightgbm.sklearn
import lale.docstrings
import lale.operators

class LGBMRegressorImpl():

    def __init__(self, boosting_type='gbdt', num_leaves=31, max_depth=-1, 
            learning_rate=0.1, n_estimators=100, subsample_for_bin=200000, 
            objective=None, class_weight=None, min_split_gain=0.0, 
            min_child_weight=0.001, min_child_samples=20, subsample=1.0, 
            subsample_freq=0, colsample_bytree=1.0, reg_alpha=0.0, reg_lambda=0.0, 
            random_state=None, n_jobs=-1, silent=True, importance_type='split'):
        self._hyperparams = {
            'boosting_type': boosting_type,
            'num_leaves': num_leaves,
            'max_depth': max_depth,
            'learning_rate': learning_rate,
            'n_estimators': n_estimators,
            'subsample_for_bin': subsample_for_bin,
            'objective': objective,
            'class_weight': class_weight,
            'min_split_gain': min_split_gain,
            'min_child_weight': min_child_weight,
            'min_child_samples': min_child_samples,
            'subsample': subsample,
            'subsample_freq': subsample_freq,
            'colsample_bytree': colsample_bytree,
            'reg_alpha': reg_alpha,
            'reg_lambda': reg_lambda,
            'random_state': random_state,
            'n_jobs': n_jobs,
            'silent': silent,
            'importance_type': importance_type
        }
        self._wrapped_model = lightgbm.sklearn.LGBMRegressor(**self._hyperparams)

    def fit(self, X, y=None, **fit_params):
        try:
            if fit_params is None:
                self._wrapped_model.fit(X, y)
            else:
                self._wrapped_model.fit(X, y, **fit_params)
        except Exception as e:
            raise RuntimeError(str(self._hyperparams)) from e
        return self

    def predict(self, X):
        return self._wrapped_model.predict(X)

    def predict_proba(self, X):
        return self._wrapped_model.predict_proba(X)

_hyperparams_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'LightGBM classifier. (https://lightgbm.readthedocs.io/en/latest/Python-API.html#scikit-learn-api)',
    'allOf': [{
        'type': 'object',
        'relevantToOptimizer': ['boosting_type', 'max_depth', 'learning_rate', 'n_estimators',
            'min_child_samples', 'subsample'],
        'additionalProperties': False,
        'properties': {
            'boosting_type': {
                'enum': ['gbdt', 'dart', 'goss', 'rf'],
                'default': 'gbdt',
                'description': '‘gbdt’, traditional Gradient Boosting Decision Tree. ‘dart’, Dropouts meet Multiple Additive Regression Trees. ‘goss’, Gradient-based One-Side Sampling. ‘rf’, Random Forest.'},
            'num_leaves': {
                'type': 'integer',
                'default': 31,
                'description': 'Maximum tree leaves for base learners'}, 
            'max_depth': {
                'anyOf': [{
                    'type': 'integer',
                    'minimumForOptimizer': 3,
                    'maximumForOptimizer': 5}],
                'default': -1,
                'description': 'Maximum tree depth for base learners, <=0 means no limit'},
            'learning_rate': {
                'type': 'number',
                'minimumForOptimizer': 0.01,
                'maximumForOptimizer': 1.0,
                'distribution': 'loguniform',
                'default': 0.1,
                'description': 'Boosting learning rate.'},
            'n_estimators': {
                'type': 'integer',
                'minimumForOptimizer': 50,
                'maximumForOptimizer': 1000,
                'distribution': 'uniform',
                'default': 100,
                'description': 'Number of boosted trees to fit.'},
            'subsample_for_bin': {
                'type': 'integer',
                'default': 200000,
                'description': 'Number of samples for constructing bins.'},
            'objective': {
                'anyOf': [{
                    'type': 'object'}, {
                    'enum': ['regression', None]}],
                'default': None,
                'description': 'Specify the learning task and the corresponding learning objective or a custom objective function to be used'},
            'class_weight': {#Should not apply to regression, but documentation includes it.
                'anyOf': [{
                    'type': 'object'}, {
                    'enum': ['balanced', None]}],
                'default': None,
                'description': 'Weights associated with classes'},
            'min_split_gain': {
                'type': 'number',
                'default': 0.0,
                'description': 'Minimum loss reduction required to make a further partition on a leaf node of the tree.'},
            'min_child_weight': {
                'type': 'number',
                'default': 1e-3,
                'description': 'Minimum sum of instance weight (hessian) needed in a child (leaf).'},
            'min_child_samples': {
                'type': 'integer',
                'minimumForOptimizer': 1,
                'maximumForOptimizer': 30,
                'distribution': 'uniform',
                'default': 20,
                'description': 'Minimum number of data needed in a child (leaf).'},
            'subsample': {
                'type': 'number',
                'minimumForOptimizer': 0.01,
                'maximumForOptimizer': 1.0,
                'distribution': 'uniform',
                'default': 1.0,
                'description': 'Subsample ratio of the training instance.'},
            'subsample_freq': {
                'type': 'integer',
                'default': 0,
                'description': 'Frequence of subsample, <=0 means no enable.'},
            'colsample_bytree': {
                'type': 'number',
                'default': 1.0,
                'description': 'Subsample ratio of columns when constructing each tree.'},
            'reg_alpha': {
                'type': 'number',
                'minimumForOptimizer': 0.0,
                'maximumForOptimizer': 1.0,
                'default': 0.0,
                'description': 'L1 regularization term on weights.'},
            'reg_lambda': {
                'type': 'number',
                'minimumForOptimizer': 0.0,
                'maximumForOptimizer': 1.0,
                'default': 0.0,
                'description': 'L2 regularization term on weights.'},
            'random_state': {
                'anyOf': [{
                    'type': 'integer'},{
                    'enum': [None]}],
                'default': None,
                'description': 'Random number seed. If None, default seeds in C++ code will be used.'},
            'n_jobs': {
                'type': 'integer',
                'default': -1,
                'description': 'Number of parallel threads.'},
            'silent': {
                'type': 'boolean',
                'default': True,
                'description': 'Whether to print messages while running boosting.'},
            'importance_type': {
                'enum': ['split', 'gain'],
                'default': 'split',
                'description': 'The type of feature importance to be filled into feature_importances_.'},
        }}]
}
_input_fit_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Build a lightgbm model from the training set (X, y).',
    'type': 'object',
    'required': ['X', 'y'],
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            },
            'description': 'The input samples. Internally, it will be converted to'},
        'y': {
            'type': 'array',
            'items': {'type': 'number'},
            'description': 'Target values real numbers'},
        'sample_weight': {
            'anyOf': [{
                'type': 'array',
                'items': {
                    'type': 'number'},
            }, {
                'enum': [None]}],
            'default': None,
            'description': 'Weights of training data.'},
        'init_score': {
            'anyOf': [{
                'type': 'array',
                'items': {
                    'type': 'number'},
            }, {
                'enum': [None]}],
            'default': None,
            'description': 'Init score of training data.'},
        'group': {
            'default': None,
            'description': 'Group data of training data.'},
        'eval_set': {
            'default': None,
            'description': 'A list of (X, y) tuple pairs to use as validation sets.'},
        'eval_names': {
            'default': None,
            'description': 'Names of eval_set.'},
        'eval_sample_weight': {
            'default': None,
            'description': 'Weights of eval data.'},
        'eval_class_weight': {
            'anyOf': [{
                'type': 'array',
                'items': {'type':'number'}}, {
                'enum': [None]}],
            'default': None,
            'description': 'Class weights of eval data.'},
        'eval_init_score': {
            'default': None,
            'description': 'Init score of eval data.'},
        'eval_group': {
            'default': None,
            'description': 'Group data of eval data.'},
        'eval_metric': {
            'anyOf': [{
                'type': 'array',
                'items': {'type':'string'}}, {
                'enum': ['l2', None]},{
                'type': 'object'}],
            'default': None,
            'description': 'string, list of strings, callable or None, optional (default=None).'},
        'early_stopping_rounds': {
            'anyOf': [{
                'type': 'integer'}, {
                'enum': [None]}],
            'default': None,
            'description': 'Activates early stopping. The model will train until the validation score stops improving.'},
        'verbose': {
            'anyOf': [{
                'type': 'boolean'}, {
                'type': 'integer'}],
            'default': True,
            'description': 'Requires at least one evaluation data.'},
        'feature_name': {
            'anyOf': [{
                'type': 'array',
                'items': {'type':'string'}}, {
                'enum': ['auto']}],
            'default': 'auto',
            'description': 'Feature names. If ‘auto’ and data is pandas DataFrame, data columns names are used.'},
        'categorical_feature': {
            'anyOf': [{
                'type': 'array',
                'items': {'anyOf': [{'type':'string'}, {'type':'integer'}]}}, {
                'enum': ['auto']}],
            'default': 'auto',
            'description': 'Categorical features. If list of int, interpreted as indices. If list of strings, interpreted as feature names.'},
        'callbacks': {
            'anyOf': [{
                'type': 'array',
                'items': {'type':'object'}}, {
                'enum': [None]}],
            'default': None,
            'description': 'List of callback functions that are applied at each iteration. '}
    }
}
_input_predict_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Return the predicted value for each sample.',
    'type': 'object',
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            },
            'description': ' Input features matrix.'},
        'raw_score': {
            'type': 'boolean',
            'default': False,
            'description': 'Whether to predict raw scores.'},
        'num_iteration': {
            'anyOf': [{
                'type': 'integer'}, {
                'enum': [None]}],
            'default': None,
            'description': 'Limit number of iterations in the prediction.'},
        'pred_leaf': {
            'type': 'boolean',
            'default': False,
            'description': 'Whether to predict leaf index.'},
        'pred_contrib': {
            'type': 'boolean',
            'default': False,
            'description': 'Whether to predict feature contributions.'},
    },
}
_output_predict_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Return the predicted value for each sample.',
    'type': 'array',
    'items': {'type': 'number'},
}
_input_predict_proba_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Return the predicted probability for each class for each sample.',
    'type': 'object',
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            },
            'description': ' Input features matrix.'},
        'raw_score': {
            'type': 'boolean',
            'default': False,
            'description': 'Whether to predict raw scores.'},
        'num_iteration': {
            'anyOf': [{
                'type': 'integer'}, {
                'enum': [None]}],
            'default': None,
            'description': 'Limit number of iterations in the prediction.'},
        'pred_leaf': {
            'type': 'boolean',
            'default': False,
            'description': 'Whether to predict leaf index.'},
        'pred_contrib': {
            'type': 'boolean',
            'default': False,
            'description': 'Whether to predict feature contributions.'},
    },
}
_output_predict_proba_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Return the predicted probability for each class for each sample.',
    'type': 'array',
    'items': {
        'type': 'array',
        'items': {
            'type': 'number'},
    }
}
_combined_schemas = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Combined schema for expected data and hyperparameters.',
    'type': 'object',
    'tags': {
        'pre': [],
        'op': ['estimator', 'regressor'],
        'post': []},
    'properties': {
        'hyperparams': _hyperparams_schema,
        'input_fit': _input_fit_schema,
        'input_predict': _input_predict_schema,
        'output_predict': _output_predict_schema,
        'input_predict_proba': _input_predict_proba_schema,
        'output_predict_proba': _output_predict_proba_schema}}

lale.docstrings.set_docstrings(LGBMRegressorImpl, _combined_schemas)

LGBMRegressor = lale.operators.make_operator(LGBMRegressorImpl, _combined_schemas)



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from setuptools import setup, find_packages
import os
import sys
import logging
logger = logging.getLogger(__name__)
logger.addHandler(logging.StreamHandler())

try:
    import builtins
    # This trick is borrowed from scikit-learn
    # This is a bit (!) hackish: we are setting a global variable so that the
    # main lale __init__ can detect if it is being loaded by the setup
    # routine, to avoid attempting to import components before installation.
    builtins.__LALE_SETUP__ = True
except ImportError:
    pass

with open("README.md", "r") as fh:
    long_description = fh.read()

on_rtd = os.environ.get('READTHEDOCS') == 'True'
if on_rtd:
    install_requires = []
else:
    install_requires=[
        'astunparse',
        'graphviz',
        'hyperopt==0.2.3',
        'jsonschema',
        'jsonsubschema',
        'numpy',
        'scikit-learn==0.20.3',
        'scipy',
        'pandas<=0.25.3',
        'xgboost',
        'lightgbm',
        'decorator',
        'h5py']

import lale
VERSION=lale.__version__

setup(
    name='lale',
    version=VERSION,
    author="Guillaume Baudart, Martin Hirzel, Kiran Kate, Parikshit Ram, Avraham Shinnar",
    description="Library for Semi-Automated Data Science",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/IBM/lale",
    python_requires='>=3.6',
    packages=find_packages(),
    license='',
    install_requires = install_requires,
    extras_require={
        'full': [
            'liac-arff>=2.4.0',
            'pytorch-pretrained-bert>=0.6.1',
            'torchvision>=0.2.2',
            'tensorflow-datasets>=1.0.1',
            'tensorflow>=1.13.1',
            'tensorflow_hub',
            'spacy',
            'smac<=0.10.0',
            'aif360',
            'numba',
            'torch>=1.0',
            'BlackBoxAuditing',
            'imbalanced-learn==0.4.3'],
        'test':[
            'autoai-libs',
            'jupyter',
            'mypy',
            'flake8',
            'numpydoc',
            'sphinx',
            'm2r',
            'sphinx_rtd_theme',
            'sphinxcontrib.apidoc'
        ]}
)



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sklearn.tree.tree
import lale.docstrings
import lale.operators

class DecisionTreeRegressorImpl():

    def __init__(self, criterion='mse', splitter='best', max_depth=None, min_samples_split=2, min_samples_leaf=1, min_weight_fraction_leaf=0.0, max_features=None, random_state=None, max_leaf_nodes=None, min_impurity_decrease=0.0, min_impurity_split=None, presort=False):
        self._hyperparams = {
            'criterion': criterion,
            'splitter': splitter,
            'max_depth': max_depth,
            'min_samples_split': min_samples_split,
            'min_samples_leaf': min_samples_leaf,
            'min_weight_fraction_leaf': min_weight_fraction_leaf,
            'max_features': max_features,
            'random_state': random_state,
            'max_leaf_nodes': max_leaf_nodes,
            'min_impurity_decrease': min_impurity_decrease,
            'min_impurity_split': min_impurity_split,
            'presort': presort}
        self._wrapped_model = sklearn.tree.tree.DecisionTreeRegressor(**self._hyperparams)

    def fit(self, X, y, **fit_params):
        if fit_params is None:
            self._wrapped_model.fit(X, y)
        else:
            self._wrapped_model.fit(X, y, **fit_params)
        return self

    def predict(self, X):
        return self._wrapped_model.predict(X)

_hyperparams_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'A decision tree regressor.',
    'allOf': [
    {   'type': 'object',
        'required': ['criterion', 'splitter', 'max_depth', 'min_samples_split', 'min_samples_leaf', 'max_features'],
        'relevantToOptimizer': ['criterion', 'splitter', 'max_depth', 'min_samples_split', 'min_samples_leaf', 'max_features'],
        'additionalProperties': False,
        'properties': {
            'criterion': {
                'description': 'Function to measure the quality of a split.',
                'enum': ['mse', 'friedman_mse', 'mae'],
                'default': 'mse'},
            'splitter': {
                'enum': ['best', 'random'],
                'default': 'best',
                'description': 'Strategy to choose the split at each node.'},
            'max_depth': {
                'description': 'Maximum depth of the tree.',
                'default': None,
                'anyOf': [
                {   'type': 'integer',
                    'minimum': 1,
                    'minimumForOptimizer': 3,
                    'maximumForOptimizer': 5},
                {   'enum': [None],
                    'description': 'If None, then nodes are expanded until all leaves are pure, or until all leaves contain less than min_samples_split samples.'}]},
            'min_samples_split': {
                'description': 'Minimum number of samples required to split an internal node.',
                'anyOf': [
                {   'type': 'integer',
                    'minimumForOptimizer': 2,
                    'maximumForOptimizer': 20,
                    'distribution': 'uniform'},
                {   'type': 'number',
                    'minimumForOptimizer': 0.01,
                    'maximumForOptimizer': 0.5}],
                'default': 2},
            'min_samples_leaf': {
                'description': 'Minimum number of samples required to be at a leaf node.',
                'anyOf': [
                {   'type': 'integer',
                    'minimumForOptimizer': 1,
                    'maximumForOptimizer': 20,
                    'distribution': 'uniform'},
                {   'type': 'number',
                    'minimumForOptimizer': 0.01,
                    'maximumForOptimizer': 0.5}],
                'default': 1},
            'min_weight_fraction_leaf': {
                'type': 'number',
                'default': 0.0,
                'description': 'Minimum weighted fraction of the sum total of weights (of all the input samples) required to be at a leaf node.'},
            'max_features': {
                'description': 'The number of features to consider when looking for the best split.',
                'anyOf': [
                {   'type': 'integer',
                    'forOptimizer': False},
                {   'type': 'number',
                    'minimum': 0.0,
                    'exclusiveMinimum': True,
                    'minimumForOptimizer': 0.0,
                    'maximumForOptimizer': 1.0,
                    'distribution': 'uniform'},
                {   'enum': ['auto', 'sqrt', 'log2', None]}],
                'default': None},
            'random_state': {
                'anyOf': [
                {   'type': 'integer'},
                {   'type': 'object'},
                {   'enum': [None]}],
                'default': None},
            'max_leaf_nodes': {
                'anyOf': [
                {   'type': 'integer'},
                {   'enum': [None]}],
                'default': None,
                'description': 'Grow a tree with `max_leaf_nodes` in best-first fashion.'},
            'min_impurity_decrease': {
                'type': 'number',
                'default': 0.0,
                'description': 'A node will be split if this split induces a decrease of the impurity greater than or equal to this value.'},
            'min_impurity_split': {
                'anyOf':[
                {'type': 'number'},{
                    'enum': [None]
                }],
                'default': None,
                'description': 'Threshold for early stopping in tree growth.'},
            'presort': {
                'type': 'boolean',
                'default': False,
                'description': 'Whether to presort the data to speed up the finding of best splits in fitting.'},
        }}]}

_input_fit_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Build a decision tree regressor from the training set (X, y).',
    'type': 'object',
    'required': ['X', 'y'],
    'properties': {
        'X': {
            'anyOf': [{
                'type': 'array',
                'items': {
                    'type': 'array',
                    'items': {
                        'type': 'number'},
                }}],
            'description': 'The training input samples. Internally, it will be converted to'},
        'y': {
            'type': 'array',
            'items': {
                'type': 'number'},
            'description': 'The target values (real numbers). Use ``dtype=np.float64`` and'},
        'sample_weight': {
            'anyOf': [{
                'type': 'array',
                'items': {
                    'type': 'number'},
            }, {
                'enum': [None]}],
            'description': 'Sample weights. If None, then samples are equally weighted. Splits'},
        'check_input': {
            'type': 'boolean',
            'default': True,
            'description': 'Allow to bypass several input checking.'},
        'X_idx_sorted': {
            'anyOf': [{
                'type': 'array',
                'items': {
                    'type': 'array',
                    'items': {
                        'type': 'number'},
                }}, {
                'enum': [None]}],
            'default': None,
            'description': 'The indexes of the sorted training input samples. If many tree'},
    },
}
_input_predict_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Predict class or regression value for X.',
    'type': 'object',
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            },
            'description': 'The input samples. Internally, its dtype will be converted to'},
        'check_input': {
            'type': 'boolean',
            'default': True,
            'description': 'Allow to bypass several input checking.'},
    },
}
_output_predict_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'The predicted classes, or the predict values.',
    'type': 'array',
    'items': {
        'type': 'number'},
}
_combined_schemas = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': """`Decision tree regressor`_ from scikit-learn.

.. _`Decision tree regressor`: https://scikit-learn.org/0.20/modules/generated/sklearn.tree.DecisionTreeRegressor.html#sklearn-tree-decisiontreeregressor
""",
    'documentation_url': 'https://lale.readthedocs.io/en/latest/modules/lale.lib.sklearn.decision_tree_regressor.html',
    'type': 'object',
    'tags': {
        'pre': [],
        'op': ['estimator', 'regressor'],
        'post': []},
    'properties': {
        'hyperparams': _hyperparams_schema,
        'input_fit': _input_fit_schema,
        'input_predict': _input_predict_schema,
        'output_predict': _output_predict_schema}}

lale.docstrings.set_docstrings(DecisionTreeRegressorImpl, _combined_schemas)

DecisionTreeRegressor = lale.operators.make_operator(DecisionTreeRegressorImpl, _combined_schemas)




from sklearn.naive_bayes import ComplementNB as Op
import lale.helpers
import lale.operators
import lale.docstrings
from numpy import nan, inf

class ComplementNBImpl():

    def __init__(self, alpha=1.0, fit_prior=True, class_prior=None, norm=False):
        self._hyperparams = {
            'alpha': alpha,
            'fit_prior': fit_prior,
            'class_prior': class_prior,
            'norm': norm}
        self._wrapped_model = Op(**self._hyperparams)

    def fit(self, X, y=None):
        if (y is not None):
            self._wrapped_model.fit(X, y)
        else:
            self._wrapped_model.fit(X)
        return self

    def predict(self, X):
        return self._wrapped_model.predict(X)

    def predict_proba(self, X):
        return self._wrapped_model.predict_proba(X)
_hyperparams_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'inherited docstring for ComplementNB    The Complement Naive Bayes classifier described in Rennie et al. (2003).',
    'allOf': [{
        'type': 'object',
        'required': ['alpha', 'fit_prior', 'class_prior', 'norm'],
        'relevantToOptimizer': [],
        'additionalProperties': False,
        'properties': {
            'alpha': {
                'type': 'number',
                'default': 1.0,
                'description': 'Additive (Laplace/Lidstone) smoothing parameter (0 for no smoothing).'},
            'fit_prior': {
                'type': 'boolean',
                'default': True,
                'description': 'Only used in edge case with a single class in the training set.'},
            'class_prior': {
                'anyOf': [{
                    'type': 'array',
                    'items': {
                        'type': 'number'},
                }, {
                    'enum': [None]}],
                'default': None,
                'description': 'Prior probabilities of the classes'},
            'norm': {
                'type': 'boolean',
                'default': False,
                'description': 'Whether or not a second normalization of the weights is performed'},
        }}, {
        'XXX TODO XXX': 'Parameter: fit_prior > only used in edge case with a single class in the training set'}],
}
_input_fit_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Fit Naive Bayes classifier according to X, y',
    'type': 'object',
    'required': ['X', 'y'],
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            },
            'description': 'Training vectors, where n_samples is the number of samples and n_features is the number of features.'},
        'y': {
            'type': 'array',
            'items': {
                'type': 'number'},
            'description': 'Target values.'},
        'sample_weight': {
            'anyOf': [{
                'type': 'array',
                'items': {
                    'type': 'number'},
            }, {
                'enum': [None]}],
            'default': None,
            'description': 'Weights applied to individual samples (1'},
    },
}
_input_predict_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Perform classification on an array of test vectors X.',
    'type': 'object',
    'required': ['X'],
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            }},
    },
}
_output_predict_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Predicted target values for X',
    'type': 'array',
    'items': {
        'type': 'number'},
}
_input_predict_proba_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Return probability estimates for the test vector X.',
    'type': 'object',
    'required': ['X'],
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            }},
    },
}
_output_predict_proba_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Returns the probability of the samples for each class in the model',
    'type': 'array',
    'items': {
        'type': 'array',
        'items': {
            'type': 'number'},
    },
}
_combined_schemas = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Combined schema for expected data and hyperparameters.',
    'documentation_url': 'https://scikit-learn.org/0.20/modules/generated/sklearn.naive_bayes.ComplementNB#sklearn-naive_bayes-complementnb',
    'type': 'object',
    'tags': {
        'pre': [],
        'op': ['estimator'],
        'post': []},
    'properties': {
        'hyperparams': _hyperparams_schema,
        'input_fit': _input_fit_schema,
        'input_predict': _input_predict_schema,
        'output_predict': _output_predict_schema,
        'input_predict_proba': _input_predict_proba_schema,
        'output_predict_proba': _output_predict_proba_schema},
}
lale.docstrings.set_docstrings(ComplementNBImpl, _combined_schemas)
ComplementNB = lale.operators.make_operator(ComplementNBImpl, _combined_schemas)





from sklearn.linear_model.ridge import RidgeClassifier as Op
import lale.helpers
import lale.operators
import lale.docstrings
from numpy import nan, inf

class RidgeClassifierImpl():

    def __init__(self, alpha=1.0, fit_intercept=True, normalize=False, copy_X=True, max_iter=None, tol=0.001, class_weight='balanced', solver='auto', random_state=None):
        self._hyperparams = {
            'alpha': alpha,
            'fit_intercept': fit_intercept,
            'normalize': normalize,
            'copy_X': copy_X,
            'max_iter': max_iter,
            'tol': tol,
            'class_weight': class_weight,
            'solver': solver,
            'random_state': random_state}
        self._wrapped_model = Op(**self._hyperparams)

    def fit(self, X, y=None):
        if (y is not None):
            self._wrapped_model.fit(X, y)
        else:
            self._wrapped_model.fit(X)
        return self

    def predict(self, X):
        return self._wrapped_model.predict(X)

    def decision_function(self, X):
        return self._wrapped_model.decision_function(X)
_hyperparams_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'inherited docstring for RidgeClassifier    Classifier using Ridge regression.',
    'allOf': [{
        'type': 'object',
        'required': ['alpha', 'fit_intercept', 'normalize', 'copy_X', 'max_iter', 'tol', 'class_weight', 'solver', 'random_state'],
        'relevantToOptimizer': ['alpha', 'fit_intercept', 'normalize', 'copy_X', 'max_iter', 'tol', 'solver'],
        'additionalProperties': False,
        'properties': {
            'alpha': {
                'type': 'number',
                'minimumForOptimizer': 1e-10,
                'maximumForOptimizer': 1.0,
                'distribution': 'loguniform',
                'default': 1.0,
                'description': 'Regularization strength; must be a positive float'},
            'fit_intercept': {
                'type': 'boolean',
                'default': True,
                'description': 'Whether to calculate the intercept for this model'},
            'normalize': {
                'type': 'boolean',
                'default': False,
                'description': 'This parameter is ignored when ``fit_intercept`` is set to False'},
            'copy_X': {
                'type': 'boolean',
                'default': True,
                'description': 'If True, X will be copied; else, it may be overwritten.'},
            'max_iter': {
                'anyOf': [{
                    'type': 'integer',
                    'minimumForOptimizer': 10,
                    'maximumForOptimizer': 1000,
                    'distribution': 'uniform'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'Maximum number of iterations for conjugate gradient solver'},
            'tol': {
                'type': 'number',
                'minimumForOptimizer': 1e-08,
                'maximumForOptimizer': 0.01,
                'distribution': 'loguniform',
                'default': 0.001,
                'description': 'Precision of the solution.'},
            'class_weight': {
                'XXX TODO XXX': "dict or 'balanced', optional",
                'description': 'Weights associated with classes in the form ``{class_label: weight}``',
                'enum': ['balanced'],
                'default': 'balanced'},
            'solver': {
                'enum': ['auto', 'svd', 'cholesky', 'lsqr', 'sparse_cg', 'sag', 'saga'],
                'default': 'auto',
                'description': "Solver to use in the computational routines:  - 'auto' chooses the solver automatically based on the type of data"},
            'random_state': {
                'anyOf': [{
                    'type': 'integer'}, {
                    'type': 'object'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'The seed of the pseudo random number generator to use when shuffling the data'},
        }}, {
        'XXX TODO XXX': 'Parameter: solver > only guaranteed on features with approximately the same scale'}],
}
_input_fit_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Fit Ridge regression model.',
    'type': 'object',
    'required': ['X', 'y'],
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            },
            'description': 'Training data'},
        'y': {
            'type': 'array',
            'items': {
                'type': 'number'},
            'description': 'Target values'},
        'sample_weight': {
            'anyOf': [{
                'type': 'number'}, {
                'type': 'array',
                'items': {
                    'type': 'number'},
            }],
            'description': 'Sample weight'},
    },
}
_input_predict_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Predict class labels for samples in X.',
    'type': 'object',
    'required': ['X'],
    'properties': {
        'X': {
            'anyOf': [{
                'type': 'array',
                'items': {
                    'laleType': 'Any',
                    'XXX TODO XXX': 'item type'},
                'XXX TODO XXX': 'array_like or sparse matrix, shape (n_samples, n_features)'}, {
                'type': 'array',
                'items': {
                    'type': 'array',
                    'items': {
                        'type': 'number'},
                }}],
            'description': 'Samples.'},
    },
}
_output_predict_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Predicted class label per sample.',
    'type': 'array',
    'items': {
        'type': 'number'},
}
_input_decision_function_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Predict confidence scores for samples.',
    'type': 'object',
    'required': ['X'],
    'properties': {
        'X': {
            'anyOf': [{
                'type': 'array',
                'items': {
                    'laleType': 'Any',
                    'XXX TODO XXX': 'item type'},
                'XXX TODO XXX': 'array_like or sparse matrix, shape (n_samples, n_features)'}, {
                'type': 'array',
                'items': {
                    'type': 'array',
                    'items': {
                        'type': 'number'},
                }}],
            'description': 'Samples.'},
    },
}
_output_decision_function_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Confidence scores per (sample, class) combination',
    'laleType': 'Any',
    'XXX TODO XXX': 'array, shape=(n_samples,) if n_classes == 2 else (n_samples, n_classes)',
}
_combined_schemas = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Combined schema for expected data and hyperparameters.',
    'documentation_url': 'https://scikit-learn.org/0.20/modules/generated/sklearn.linear_model.RidgeClassifier#sklearn-linear_model-ridgeclassifier',
    'type': 'object',
    'tags': {
        'pre': [],
        'op': ['estimator', 'classifier'],
        'post': []},
    'properties': {
        'hyperparams': _hyperparams_schema,
        'input_fit': _input_fit_schema,
        'input_predict': _input_predict_schema,
        'output_predict': _output_predict_schema,
        'input_decision_function': _input_decision_function_schema,
        'output_decision_function': _output_decision_function_schema},
}
lale.docstrings.set_docstrings(RidgeClassifierImpl, _combined_schemas)
RidgeClassifier = lale.operators.make_operator(RidgeClassifierImpl, _combined_schemas)




# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sklearn.ensemble.forest
import lale.docstrings
import lale.operators

class ExtraTreesRegressorImpl():

    def __init__(self, n_estimators=10, criterion='mse', max_depth=None, min_samples_split=2, min_samples_leaf=1, min_weight_fraction_leaf=0.0, max_features='auto', max_leaf_nodes=None, min_impurity_decrease=0.0, min_impurity_split=None, bootstrap=False, oob_score=False, n_jobs=None, random_state=None, verbose=0, warm_start=False):
        self._hyperparams = {
            'n_estimators': n_estimators,
            'criterion': criterion,
            'max_depth': max_depth,
            'min_samples_split': min_samples_split,
            'min_samples_leaf': min_samples_leaf,
            'min_weight_fraction_leaf': min_weight_fraction_leaf,
            'max_features': max_features,
            'max_leaf_nodes': max_leaf_nodes,
            'min_impurity_decrease': min_impurity_decrease,
            'min_impurity_split': min_impurity_split,
            'bootstrap': bootstrap,
            'oob_score': oob_score,
            'n_jobs': n_jobs,
            'random_state': random_state,
            'verbose': verbose,
            'warm_start': warm_start}
        self._wrapped_model = sklearn.ensemble.forest.ExtraTreesRegressor(**self._hyperparams)

    def fit(self, X, y, **fit_params):
        if fit_params is None:
            self._wrapped_model.fit(X, y)
        else:
            self._wrapped_model.fit(X, y, **fit_params)
        return self

    def predict(self, X):
        return self._wrapped_model.predict(X)

_hyperparams_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'An extra-trees regressor.',
    'allOf': [{
        'type': 'object',
        'required': ['n_estimators', 'criterion', 'max_depth', 'min_samples_split', 'min_samples_leaf', 'max_features', 'bootstrap'],
        'relevantToOptimizer': ['n_estimators', 'criterion', 'max_depth', 'min_samples_split', 'min_samples_leaf', 'max_features', 'bootstrap'],
        'additionalProperties': False,
        'properties': {
            'n_estimators': {
                'type': 'integer',
                'minimumForOptimizer': 10,
                'maximumForOptimizer': 100,
                'default': 10,
                'description': 'The number of trees in the forest.'},
            'criterion': {
                'enum': ['mae', 'mse', 'friedman_mse'],
                'default': 'mse',
                'description': 'The function to measure the quality of a split. Supported criteria'},
            'max_depth': {
                'anyOf': [{
                    'type': 'integer',
                    'minimumForOptimizer': 3,
                    'maximumForOptimizer': 5}, {
                    'enum': [None]}],
                'default': None,
                'description': 'The maximum depth of the tree. If None, then nodes are expanded until'},
            'min_samples_split': {
                'anyOf': [{
                    'type': 'integer',
                    'minimumForOptimizer': 2,
                    'maximumForOptimizer': 20,
                    'distribution': 'uniform'}, {
                    'type': 'number',
                    'minimumForOptimizer': 0.01,
                    'maximumForOptimizer': 0.5}],
                'default': 2,
                'description': 'The minimum number of samples required to split an internal node:'},
            'min_samples_leaf': {
                'anyOf': [{
                    'type': 'integer',
                    'minimumForOptimizer': 1,
                    'maximumForOptimizer': 20,
                    'distribution': 'uniform'}, {
                    'type': 'number',
                    'minimumForOptimizer': 0.01,
                    'maximumForOptimizer': 0.5}],
                'default': 1,
                'description': 'The minimum number of samples required to be at a leaf node.'},
            'min_weight_fraction_leaf': {
                'type': 'number',
                'default': 0.0,
                'description': 'The minimum weighted fraction of the sum total of weights (of all'},
            'max_features': {
                'anyOf': [{
                    'type': 'integer',
                    'forOptimizer': False}, {
                    'type': 'number',
                    'minimum': 0.0,
                    'exclusiveMinimum': True,
                    'minimumForOptimizer': 0.0,
                    'maximumForOptimizer': 1.0,
                    'distribution': 'uniform'}, {
                    'enum': ['auto', 'sqrt', 'log2', None]}],
                'default': 'auto',
                'description': 'The number of features to consider when looking for the best split:'},
            'max_leaf_nodes': {
                'anyOf': [{
                    'type': 'integer'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'Grow trees with ``max_leaf_nodes`` in best-first fashion.'},
            'min_impurity_decrease': {
                'type': 'number',
                'default': 0.0,
                'description': 'A node will be split if this split induces a decrease of the impurity'},
            'min_impurity_split': {
                'anyOf':[
                {'type': 'number'},{
                    'enum': [None]
                }],
                'default': None,
                'description': 'Threshold for early stopping in tree growth. A node will split'},
            'bootstrap': {
                'type': 'boolean',
                'default': False,
                'description': 'Whether bootstrap samples are used when building trees. If False, the'},
            'oob_score': {
                'type': 'boolean',
                'default': False,
                'description': 'Whether to use out-of-bag samples to estimate the R^2 on unseen data.'},
            'n_jobs': {
                'anyOf': [{
                    'type': 'integer'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'The number of jobs to run in parallel for both `fit` and `predict`.'},
            'random_state': {
                'anyOf': [{
                    'type': 'integer'}, {
                    'type': 'object'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'If int, random_state is the seed used by the random number generator;'},
            'verbose': {
                'type': 'integer',
                'default': 0,
                'description': 'Controls the verbosity when fitting and predicting.'},
            'warm_start': {
                'type': 'boolean',
                'default': False,
                'description': 'When set to ``True``, reuse the solution of the previous call to fit'},
        }}]
}
_input_fit_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Build a forest of trees from the training set (X, y).',
    'type': 'object',
    'required': ['X', 'y'],
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            },
            'description': 'The training input samples. Internally, its dtype will be converted'},
        'y': {
            'type': 'array',
            'items': {
                'type': 'number'},
            'description': 'The target values (class labels in classification, real numbers in'},
        'sample_weight': {
            'anyOf': [{
                'type': 'array',
                'items': {
                    'type': 'number'},
            }, {
                'enum': [None]}],
            'description': 'Sample weights. If None, then samples are equally weighted. Splits'},
    },
}
_input_predict_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Predict regression target for X.',
    'type': 'object',
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            },
            'description': 'The input samples. Internally, its dtype will be converted to'},
    },
}
_output_predict_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'The predicted values.',
    'type': 'array',
    'items': {
        'type': 'number'},
}
_combined_schemas = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': """`Extra trees regressor`_ random forest from scikit-learn.

.. _`Extra trees regressor`: https://scikit-learn.org/0.20/modules/generated/sklearn.ensemble.ExtraTreesRegressor.html#sklearn-ensemble-extratreesregressor
""",
    'documentation_url': 'https://lale.readthedocs.io/en/latest/modules/lale.lib.sklearn.extra_trees_regressor.html',
    'type': 'object',
    'tags': {
        'pre': [],
        'op': ['estimator', 'regressor'],
        'post': []},
    'properties': {
        'hyperparams': _hyperparams_schema,
        'input_fit': _input_fit_schema,
        'input_predict': _input_predict_schema,
        'output_predict': _output_predict_schema}}

lale.docstrings.set_docstrings(ExtraTreesRegressorImpl, _combined_schemas)

ExtraTreesRegressor = lale.operators.make_operator(ExtraTreesRegressorImpl, _combined_schemas)




from sklearn.kernel_ridge import KernelRidge as Op
import lale.helpers
import lale.operators
import lale.docstrings
from numpy import nan, inf

class KernelRidgeImpl():

    def __init__(self, alpha=1, kernel='linear', gamma=None, degree=3, coef0=1, kernel_params=None):
        self._hyperparams = {
            'alpha': alpha,
            'kernel': kernel,
            'gamma': gamma,
            'degree': degree,
            'coef0': coef0,
            'kernel_params': kernel_params}
        self._wrapped_model = Op(**self._hyperparams)

    def fit(self, X, y=None):
        if (y is not None):
            self._wrapped_model.fit(X, y)
        else:
            self._wrapped_model.fit(X)
        return self

    def predict(self, X):
        return self._wrapped_model.predict(X)
_hyperparams_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'inherited docstring for KernelRidge    Kernel ridge regression.',
    'allOf': [{
        'type': 'object',
        'required': ['alpha', 'kernel', 'gamma', 'degree', 'coef0', 'kernel_params'],
        'relevantToOptimizer': ['alpha', 'kernel', 'degree', 'coef0'],
        'additionalProperties': False,
        'properties': {
            'alpha': {
                'XXX TODO XXX': '{float, array-like}, shape = [n_targets]',
                'description': 'Small positive values of alpha improve the conditioning of the problem and reduce the variance of the estimates',
                'type': 'integer',
                'minimumForOptimizer': 1,
                'maximumForOptimizer': 2,
                'distribution': 'uniform',
                'default': 1},
            'kernel': {
                'anyOf': [{
                    'type': 'object',
                    'forOptimizer': False}, {
                    'enum': ['linear', 'poly', 'rbf', 'sigmoid']}],
                'default': 'linear',
                'description': 'Kernel mapping used internally'},
            'gamma': {
                'anyOf': [{
                    'type': 'number'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'Gamma parameter for the RBF, laplacian, polynomial, exponential chi2 and sigmoid kernels'},
            'degree': {
                'type': 'number',
                'minimumForOptimizer': 0.0,
                'maximumForOptimizer': 1.0,
                'distribution': 'uniform',
                'default': 3,
                'description': 'Degree of the polynomial kernel'},
            'coef0': {
                'type': 'number',
                'minimumForOptimizer': 0.0,
                'maximumForOptimizer': 1.0,
                'distribution': 'uniform',
                'default': 1,
                'description': 'Zero coefficient for polynomial and sigmoid kernels'},
            'kernel_params': {
                'XXX TODO XXX': 'mapping of string to any, optional',
                'description': 'Additional parameters (keyword arguments) for kernel function passed as callable object.',
                'enum': [None],
                'default': None},
        }}],
}
_input_fit_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Fit Kernel Ridge regression model',
    'type': 'object',
    'required': ['X', 'y'],
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            },
            'description': 'Training data'},
        'y': {
            'anyOf': [{
                'type': 'array',
                'items': {
                    'type': 'number'},
            }, {
                'type': 'array',
                'items': {
                    'type': 'array',
                    'items': {
                        'type': 'number'},
                }}],
            'description': 'Target values'},
        'sample_weight': {
            'anyOf': [{
                'type': 'number'}, {
                'type': 'array',
                'items': {
                    'type': 'number'},
            }],
            'description': 'Individual weights for each sample, ignored if None is passed.'},
    },
}
_input_predict_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Predict using the kernel ridge model',
    'type': 'object',
    'required': ['X'],
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            },
            'description': 'Samples'},
    },
}
_output_predict_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Returns predicted values.',
    'anyOf': [{
        'type': 'array',
        'items': {
            'type': 'number'},
    }, {
        'type': 'array',
        'items': {
            'type': 'array',
            'items': {
                'type': 'number'},
        }}],
}
_combined_schemas = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Combined schema for expected data and hyperparameters.',
    'documentation_url': 'https://scikit-learn.org/0.20/modules/generated/sklearn.kernel_ridge.KernelRidge#sklearn-kernel_ridge-kernelridge',
    'type': 'object',
    'tags': {
        'pre': [],
        'op': ['estimator'],
        'post': []},
    'properties': {
        'hyperparams': _hyperparams_schema,
        'input_fit': _input_fit_schema,
        'input_predict': _input_predict_schema,
        'output_predict': _output_predict_schema},
}
lale.docstrings.set_docstrings(KernelRidgeImpl, _combined_schemas)
KernelRidge = lale.operators.make_operator(KernelRidgeImpl, _combined_schemas)





from sklearn.decomposition.kernel_pca import KernelPCA as Op
import lale.helpers
import lale.operators
import lale.docstrings
from numpy import nan, inf

class KernelPCAImpl():

    def __init__(self, n_components=None, kernel='linear', gamma=None, degree=3, coef0=1, kernel_params=None, alpha=1.0, fit_inverse_transform=False, eigen_solver='auto', tol=0, max_iter=None, remove_zero_eig=False, random_state=None, copy_X=True, n_jobs=None):
        self._hyperparams = {
            'n_components': n_components,
            'kernel': kernel,
            'gamma': gamma,
            'degree': degree,
            'coef0': coef0,
            'kernel_params': kernel_params,
            'alpha': alpha,
            'fit_inverse_transform': fit_inverse_transform,
            'eigen_solver': eigen_solver,
            'tol': tol,
            'max_iter': max_iter,
            'remove_zero_eig': remove_zero_eig,
            'random_state': random_state,
            'copy_X': copy_X,
            'n_jobs': n_jobs}
        self._wrapped_model = Op(**self._hyperparams)

    def fit(self, X, y=None):
        if (y is not None):
            self._wrapped_model.fit(X, y)
        else:
            self._wrapped_model.fit(X)
        return self

    def transform(self, X):
        return self._wrapped_model.transform(X)
_hyperparams_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'inherited docstring for KernelPCA    Kernel Principal component analysis (KPCA)',
    'allOf': [{
        'type': 'object',
        'required': ['n_components', 'kernel', 'gamma', 'degree', 'coef0', 'kernel_params', 'alpha', 'fit_inverse_transform', 'eigen_solver', 'tol', 'max_iter', 'remove_zero_eig', 'random_state', 'copy_X', 'n_jobs'],
        'relevantToOptimizer': ['n_components', 'kernel', 'degree', 'coef0', 'alpha', 'eigen_solver', 'tol', 'max_iter', 'remove_zero_eig', 'copy_X'],
        'additionalProperties': False,
        'properties': {
            'n_components': {
                'anyOf': [{
                    'type': 'integer',
                    'minimumForOptimizer': 2,
                    'maximumForOptimizer': 256,
                    'distribution': 'uniform'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'Number of components'},
            'kernel': {
                'enum': ['linear', 'poly', 'rbf', 'sigmoid', 'cosine', 'precomputed'],
                'default': 'linear',
                'description': 'Kernel'},
            'gamma': {
                'XXX TODO XXX': 'float, default=1/n_features',
                'description': 'Kernel coefficient for rbf, poly and sigmoid kernels',
                'enum': [None],
                'default': None},
            'degree': {
                'type': 'integer',
                'minimumForOptimizer': 2,
                'maximumForOptimizer': 3,
                'distribution': 'uniform',
                'default': 3,
                'description': 'Degree for poly kernels'},
            'coef0': {
                'type': 'number',
                'minimumForOptimizer': 0.0,
                'maximumForOptimizer': 1.0,
                'distribution': 'uniform',
                'default': 1,
                'description': 'Independent term in poly and sigmoid kernels'},
            'kernel_params': {
                'XXX TODO XXX': 'mapping of string to any, default=None',
                'description': 'Parameters (keyword arguments) and values for kernel passed as callable object',
                'enum': [None],
                'default': None},
            'alpha': {
                'anyOf': [{
                    'type': 'integer',
                    'forOptimizer': False}, {
                    'type': 'number',
                    'minimumForOptimizer': 1e-10,
                    'maximumForOptimizer': 1.0,
                    'distribution': 'loguniform'}],
                'default': 1.0,
                'description': 'Hyperparameter of the ridge regression that learns the inverse transform (when fit_inverse_transform=True).'},
            'fit_inverse_transform': {
                'type': 'boolean',
                'default': False,
                'description': 'Learn the inverse transform for non-precomputed kernels'},
            'eigen_solver': {
                'enum': ['auto', 'dense', 'arpack'],
                'default': 'auto',
                'description': 'Select eigensolver to use'},
            'tol': {
                'type': 'number',
                'minimumForOptimizer': 1e-08,
                'maximumForOptimizer': 0.01,
                'distribution': 'loguniform',
                'default': 0,
                'description': 'Convergence tolerance for arpack'},
            'max_iter': {
                'anyOf': [{
                    'type': 'integer',
                    'minimumForOptimizer': 10,
                    'maximumForOptimizer': 1000,
                    'distribution': 'uniform'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'Maximum number of iterations for arpack'},
            'remove_zero_eig': {
                'type': 'boolean',
                'default': False,
                'description': 'If True, then all components with zero eigenvalues are removed, so that the number of components in the output may be < n_components (and sometimes even zero due to numerical instability)'},
            'random_state': {
                'anyOf': [{
                    'type': 'integer'}, {
                    'type': 'object'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'If int, random_state is the seed used by the random number generator; If RandomState instance, random_state is the random number generator; If None, the random number generator is the RandomState instance used by `np.random`'},
            'copy_X': {
                'type': 'boolean',
                'default': True,
                'description': 'If True, input X is copied and stored by the model in the `X_fit_` attribute'},
            'n_jobs': {
                'anyOf': [{
                    'type': 'integer'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'The number of parallel jobs to run'},
        }}],
}
_input_fit_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Fit the model from data in X.',
    'type': 'object',
    'required': ['X'],
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            },
            'description': 'Training vector, where n_samples in the number of samples and n_features is the number of features.'},
    },
}
_input_transform_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Transform X.',
    'type': 'object',
    'required': ['X'],
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            }},
    },
}
_output_transform_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Transform X.',
    'type': 'array',
    'items': {
        'type': 'array',
        'items': {
            'type': 'number'},
    },
}
_combined_schemas = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Combined schema for expected data and hyperparameters.',
    'documentation_url': 'https://scikit-learn.org/0.20/modules/generated/sklearn.decomposition.KernelPCA#sklearn-decomposition-kernelpca',
    'type': 'object',
    'tags': {
        'pre': [],
        'op': ['transformer'],
        'post': []},
    'properties': {
        'hyperparams': _hyperparams_schema,
        'input_fit': _input_fit_schema,
        'input_transform': _input_transform_schema,
        'output_transform': _output_transform_schema},
}
lale.docstrings.set_docstrings(KernelPCAImpl, _combined_schemas)
KernelPCA = lale.operators.make_operator(KernelPCAImpl, _combined_schemas)





from sklearn.linear_model.logistic import LogisticRegressionCV as Op
import lale.helpers
import lale.operators
import lale.docstrings
from numpy import nan, inf

class LogisticRegressionCVImpl():

    def __init__(self, Cs=10, fit_intercept=True, cv=3, dual=False, penalty='l2', scoring=None, solver='lbfgs', tol=0.0001, max_iter=100, class_weight='balanced', n_jobs=None, verbose=0, refit=True, intercept_scaling=1.0, multi_class='ovr', random_state=None):
        self._hyperparams = {
            'Cs': Cs,
            'fit_intercept': fit_intercept,
            'cv': cv,
            'dual': dual,
            'penalty': penalty,
            'scoring': scoring,
            'solver': solver,
            'tol': tol,
            'max_iter': max_iter,
            'class_weight': class_weight,
            'n_jobs': n_jobs,
            'verbose': verbose,
            'refit': refit,
            'intercept_scaling': intercept_scaling,
            'multi_class': multi_class,
            'random_state': random_state}
        self._wrapped_model = Op(**self._hyperparams)

    def fit(self, X, y=None):
        if (y is not None):
            self._wrapped_model.fit(X, y)
        else:
            self._wrapped_model.fit(X)
        return self

    def predict(self, X):
        return self._wrapped_model.predict(X)

    def predict_proba(self, X):
        return self._wrapped_model.predict_proba(X)

    def decision_function(self, X):
        return self._wrapped_model.decision_function(X)
_hyperparams_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'inherited docstring for LogisticRegressionCV    Logistic Regression CV (aka logit, MaxEnt) classifier.',
    'allOf': [{
        'type': 'object',
        'required': ['Cs', 'fit_intercept', 'cv', 'dual', 'penalty', 'scoring', 'solver', 'tol', 'max_iter', 'class_weight', 'n_jobs', 'verbose', 'refit', 'intercept_scaling', 'multi_class', 'random_state'],
        'relevantToOptimizer': ['Cs', 'fit_intercept', 'cv', 'dual', 'penalty', 'scoring', 'solver', 'tol', 'max_iter', 'multi_class'],
        'additionalProperties': False,
        'properties': {
            'Cs': {
                'XXX TODO XXX': 'list of floats | int',
                'description': 'Each of the values in Cs describes the inverse of regularization strength',
                'type': 'integer',
                'minimumForOptimizer': 10,
                'maximumForOptimizer': 11,
                'distribution': 'uniform',
                'default': 10},
            'fit_intercept': {
                'type': 'boolean',
                'default': True,
                'description': 'Specifies if a constant (a.k.a'},
            'cv': {
                'XXX TODO XXX': 'integer or cross-validation generator, default: None',
                'description': 'The default cross-validation generator used is Stratified K-Folds',
                'type': 'integer',
                'minimumForOptimizer': 3,
                'maximumForOptimizer': 4,
                'distribution': 'uniform',
                'default': 3},
            'dual': {
                'type': 'boolean',
                'default': False,
                'description': 'Dual or primal formulation'},
            'penalty': {
                'XXX TODO XXX': "str, 'l1' or 'l2'",
                'description': 'Used to specify the norm used in the penalization',
                'enum': ['l2'],
                'default': 'l2'},
            'scoring': {
                'anyOf': [{
                    'type': 'object',
                    'forOptimizer': False}, {
                    'enum': ['accuracy', None]}],
                'default': None,
                'description': 'A string (see model evaluation documentation) or a scorer callable object / function with signature ``scorer(estimator, X, y)``'},
            'solver': {
                'enum': ['newton-cg', 'lbfgs', 'liblinear', 'sag', 'saga'],
                'default': 'lbfgs',
                'description': 'Algorithm to use in the optimization problem'},
            'tol': {
                'type': 'number',
                'minimumForOptimizer': 1e-08,
                'maximumForOptimizer': 0.01,
                'distribution': 'loguniform',
                'default': 0.0001,
                'description': 'Tolerance for stopping criteria.'},
            'max_iter': {
                'type': 'integer',
                'minimumForOptimizer': 10,
                'maximumForOptimizer': 1000,
                'distribution': 'uniform',
                'default': 100,
                'description': 'Maximum number of iterations of the optimization algorithm.'},
            'class_weight': {
                'XXX TODO XXX': "dict or 'balanced', optional",
                'description': 'Weights associated with classes in the form ``{class_label: weight}``',
                'enum': ['balanced'],
                'default': 'balanced'},
            'n_jobs': {
                'anyOf': [{
                    'type': 'integer'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'Number of CPU cores used during the cross-validation loop'},
            'verbose': {
                'type': 'integer',
                'default': 0,
                'description': "For the 'liblinear', 'sag' and 'lbfgs' solvers set verbose to any positive number for verbosity."},
            'refit': {
                'type': 'boolean',
                'default': True,
                'description': 'If set to True, the scores are averaged across all folds, and the coefs and the C that corresponds to the best score is taken, and a final refit is done using these parameters'},
            'intercept_scaling': {
                'type': 'number',
                'default': 1.0,
                'description': "Useful only when the solver 'liblinear' is used and self.fit_intercept is set to True"},
            'multi_class': {
                'enum': ['ovr', 'multinomial', 'auto'],
                'default': 'ovr',
                'description': "If the option chosen is 'ovr', then a binary problem is fit for each label"},
            'random_state': {
                'anyOf': [{
                    'type': 'integer'}, {
                    'type': 'object'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'If int, random_state is the seed used by the random number generator; If RandomState instance, random_state is the random number generator; If None, the random number generator is the RandomState instance used by `np.random`.'},
        }}, {
        'XXX TODO XXX': 'Parameter: dual > only implemented for l2 penalty with liblinear solver'}, {
        'XXX TODO XXX': 'Parameter: penalty > only l2 penalties'}, {
        'XXX TODO XXX': "Parameter: solver > only 'newton-cg', 'sag', 'saga' and 'lbfgs' handle multinomial loss; 'liblinear' is limited to one-versus-rest schemes"}, {
        'XXX TODO XXX': "Parameter: intercept_scaling > only when the solver 'liblinear' is used and self"}],
}
_input_fit_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Fit the model according to the given training data.',
    'type': 'object',
    'required': ['X', 'y'],
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            },
            'description': 'Training vector, where n_samples is the number of samples and n_features is the number of features.'},
        'y': {
            'type': 'array',
            'items': {
                'type': 'number'},
            'description': 'Target vector relative to X.'},
        'sample_weight': {
            'laleType': 'Any',
            'XXX TODO XXX': 'array-like, shape (n_samples,) optional',
            'description': 'Array of weights that are assigned to individual samples'},
    },
}
_input_predict_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Predict class labels for samples in X.',
    'type': 'object',
    'required': ['X'],
    'properties': {
        'X': {
            'anyOf': [{
                'type': 'array',
                'items': {
                    'laleType': 'Any',
                    'XXX TODO XXX': 'item type'},
                'XXX TODO XXX': 'array_like or sparse matrix, shape (n_samples, n_features)'}, {
                'type': 'array',
                'items': {
                    'type': 'array',
                    'items': {
                        'type': 'number'},
                }}],
            'description': 'Samples.'},
    },
}
_output_predict_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Predicted class label per sample.',
    'type': 'array',
    'items': {
        'type': 'number'},
}
_input_predict_proba_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Probability estimates.',
    'type': 'object',
    'required': ['X'],
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            }},
    },
}
_output_predict_proba_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Returns the probability of the sample for each class in the model, where classes are ordered as they are in ``self.classes_``.',
    'type': 'array',
    'items': {
        'type': 'array',
        'items': {
            'type': 'number'},
    },
}
_input_decision_function_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Predict confidence scores for samples.',
    'type': 'object',
    'required': ['X'],
    'properties': {
        'X': {
            'anyOf': [{
                'type': 'array',
                'items': {
                    'laleType': 'Any',
                    'XXX TODO XXX': 'item type'},
                'XXX TODO XXX': 'array_like or sparse matrix, shape (n_samples, n_features)'}, {
                'type': 'array',
                'items': {
                    'type': 'array',
                    'items': {
                        'type': 'number'},
                }}],
            'description': 'Samples.'},
    },
}
_output_decision_function_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Confidence scores per (sample, class) combination',
    'laleType': 'Any',
    'XXX TODO XXX': 'array, shape=(n_samples,) if n_classes == 2 else (n_samples, n_classes)',
}
_combined_schemas = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Combined schema for expected data and hyperparameters.',
    'documentation_url': 'https://scikit-learn.org/0.20/modules/generated/sklearn.linear_model.LogisticRegressionCV#sklearn-linear_model-logisticregressioncv',
    'type': 'object',
    'tags': {
        'pre': [],
        'op': ['estimator'],
        'post': []},
    'properties': {
        'hyperparams': _hyperparams_schema,
        'input_fit': _input_fit_schema,
        'input_predict': _input_predict_schema,
        'output_predict': _output_predict_schema,
        'input_predict_proba': _input_predict_proba_schema,
        'output_predict_proba': _output_predict_proba_schema,
        'input_decision_function': _input_decision_function_schema,
        'output_decision_function': _output_decision_function_schema},
}
lale.docstrings.set_docstrings(LogisticRegressionCVImpl, _combined_schemas)
LogisticRegressionCV = lale.operators.make_operator(LogisticRegressionCVImpl, _combined_schemas)

